from typing import List, Optional

from trustwise.sdk.metrics.base import BaseMetric
from trustwise.sdk.types import SensitivityRequest, SensitivityResponse


class SensitivityMetric:
    """Sensitivity metric for evaluating response sensitivity."""
    def __init__(self, client) -> None:
        self.client = client
        self.base_url = client.config.get_alignment_url("v1")

    def evaluate(
        self,
        *,
        response: Optional[str] = None,
        topics: Optional[list] = None,
        query: Optional[str] = None,
        **kwargs
    ) -> SensitivityResponse:
        """
        Evaluate the sensitivity of a response.

        Args:
            response: The response string (required)
            topics: List of topics to check for sensitivity (required)
            query: The input query string (optional)

        Returns:
            SensitivityResponse containing the evaluation results
        """
        req = BaseMetric.validate_request_model(SensitivityRequest, response=response, topics=topics, query=query, **kwargs)
        result = self.client._post(
            endpoint=f"{self.base_url}/sensitivity",
            data=req.to_dict()
        )
        return SensitivityResponse(**result)

    def batch_evaluate(
        self,
        inputs: List[SensitivityRequest]
    ) -> List[SensitivityResponse]:
        """Evaluate multiple inputs for sensitivity."""
        raise NotImplementedError("Batch evaluation not yet supported")

    def explain(
        self,
        *,
        response: str,
        topics: List[str],
        query: Optional[str] = None,
        **kwargs
    ) -> dict:
        """Get detailed explanation of the sensitivity evaluation."""
        # req = SensitivityRequest(response=response, topics=topics, query=query)
        raise NotImplementedError("Explanation not yet supported") 


from trustwise.sdk.metrics.base import BasePerformanceMetric
from trustwise.sdk.metrics.performance.v1.metrics.carbon import CarbonMetric
from trustwise.sdk.metrics.performance.v1.metrics.cost import CostMetric


class PerformanceMetricsV1(BasePerformanceMetric):
    """Performance metrics for version 1 of the API."""
    
    def __init__(self, client) -> None:
        """Initialize the performance metrics."""
        base_url = client.config.get_performance_url("v1")
        super().__init__(client, base_url=base_url, version="v1")
        self.cost = CostMetric(client)
        self.carbon = CarbonMetric(client)


from trustwise.sdk.metrics.safety.v3 import SafetyMetricsV3


class SafetyMetrics:
    """
    Namespace for Safety Metrics API versions.
    """

    def __init__(self, client) -> None:
        """
        Initialize the Safety Namespace with all supported versions.
        """
        self._current_version = SafetyMetricsV3(client)
        self.v3 = self._current_version

        # Expose all v3 methods directly
        self.faithfulness = self._current_version.faithfulness
        self.answer_relevancy = self._current_version.answer_relevancy
        self.context_relevancy = self._current_version.context_relevancy
        self.summarization = self._current_version.summarization
        self.pii = self._current_version.pii
        self.prompt_injection = self._current_version.prompt_injection
    
    @property
    def version(self) -> str:
        """Get the current default version."""
        return self._current_version.version



"""Feature flags for Trustwise SDK."""

from typing import Set

# Set of features that are currently in beta
BETA_FEATURES: Set[str] = {
    "guardrails",
}

def is_beta_feature(feature_name: str) -> bool:
    """
    Check if a feature is currently in beta.
    
    Args:
        feature_name: Name of the feature to check
        
    Returns:
        bool: True if the feature is in beta, False otherwise
    """
    return feature_name.lower() in BETA_FEATURES

def get_beta_features() -> Set[str]:
    """
    Get the set of all features currently in beta.
    
    Returns:
        Set[str]: Set of feature names that are in beta
    """
    return BETA_FEATURES.copy() 


import json

import pytest
from pydantic import ValidationError

from trustwise.sdk.types import (
    AnswerRelevancyResponse,
    ClarityResponse,
    ContextRelevancyResponse,
    Fact,
    FaithfulnessResponse,
    FormalityResponse,
    HelpfulnessResponse,
    PIIEntity,
    PIIResponse,
    PromptInjectionResponse,
    SensitivityResponse,
    SimplicityResponse,
    SummarizationResponse,
    ToneResponse,
)


def test_faithfulness_response_to_json():
    resp = FaithfulnessResponse(
        score=95.0,
        facts=[Fact(statement="Paris is the capital of France.", label="CORRECT", prob=0.99, sentence_span=[0, 30])]
    )
    # Attribute access
    assert resp.score == 95.0
    assert resp.facts[0].statement == "Paris is the capital of France."
    # JSON output
    json_str = resp.to_json()
    data = json.loads(json_str)
    assert data["score"] == 95.0
    assert data["facts"][0]["statement"] == "Paris is the capital of France."

def test_prompt_injection_response_to_json():
    resp = PromptInjectionResponse(
        score=98.0
    )
    assert resp.score == 98.0
    json_str = resp.to_json()
    data = json.loads(json_str)
    assert data["score"] == 98.0

def test_pii_response_to_json():
    resp = PIIResponse(
        identified_pii=[
            PIIEntity(interval=[0, 5], string="Hello", category="blocklist")
        ]
    )
    pii = resp.identified_pii[0]
    assert pii.interval == [0, 5]
    assert pii.string == "Hello"
    assert pii.category == "blocklist"
    json_str = resp.to_json()
    data = json.loads(json_str)
    assert data["identified_pii"][0]["interval"] == [0, 5]
    assert data["identified_pii"][0]["string"] == "Hello"
    assert data["identified_pii"][0]["category"] == "blocklist"

def test_clarity_response_to_json():
    resp = ClarityResponse(score=58.42)
    # Attribute access
    assert resp.score == 58.42
    # JSON output
    json_str = resp.to_json()
    data = json.loads(json_str)
    assert data["score"] == 58.42 

def test_score_validation_errors():
    # Single score fields
    with pytest.raises(ValidationError):
        FaithfulnessResponse(score=-1, facts=[])
    with pytest.raises(ValidationError):
        FaithfulnessResponse(score=101, facts=[])
    with pytest.raises(ValidationError):
        AnswerRelevancyResponse(score=101, generated_question="foo")
    with pytest.raises(ValidationError):
        ContextRelevancyResponse(score=101, topics=[], scores=[])
    with pytest.raises(ValidationError):
        SummarizationResponse(score=-0.1)
    with pytest.raises(ValidationError):
        PromptInjectionResponse(score=200)
    with pytest.raises(ValidationError):
        ClarityResponse(score=-5)
    with pytest.raises(ValidationError):
        HelpfulnessResponse(score=1000)
    with pytest.raises(ValidationError):
        FormalityResponse(score=50, sentences=[], scores=[-1])
    with pytest.raises(ValidationError):
        SimplicityResponse(score=101)
    # Dict/list score fields
    with pytest.raises(ValidationError):
        SensitivityResponse(scores={"topic": 101})
    with pytest.raises(ValidationError):
        ToneResponse(labels=["a"], scores=[-1]) 


from typing import Any
from unittest.mock import patch

import pytest

from trustwise.sdk import TrustwiseSDK
from trustwise.sdk.config import TrustwiseConfig
from trustwise.sdk.types import FaithfulnessResponse

from .helpers import get_mock_response


class TestTrustwiseSDK:
    """Test suite for the main Trustwise SDK class."""

    def test_sdk_initialization(self, api_key: str) -> None:
        """Test SDK initialization."""
        config = TrustwiseConfig(api_key=api_key)
        sdk = TrustwiseSDK(config)
        assert isinstance(sdk, TrustwiseSDK)
        assert hasattr(sdk, "client")
        assert hasattr(sdk, "safety")
        assert hasattr(sdk, "alignment")
        assert hasattr(sdk, "performance")

    def test_get_versions(self, sdk: TrustwiseSDK) -> None:
        """Test version information retrieval."""
        versions = sdk.get_versions()
        
        assert isinstance(versions, dict)
        assert "safety" in versions
        assert "alignment" in versions
        assert "performance" in versions
        assert isinstance(versions["safety"], list)
        assert isinstance(versions["alignment"], list)
        assert isinstance(versions["performance"], list)
        assert "v3" in versions["safety"]
        assert "v1" in versions["alignment"]
        assert "v1" in versions["performance"]

    def test_safety_namespace(self, sdk: TrustwiseSDK) -> None:
        """Test safety namespace access."""
        # Test direct version access
        assert hasattr(sdk.safety, "v3")
        assert hasattr(sdk.safety.v3, "faithfulness")
        assert hasattr(sdk.safety.v3, "answer_relevancy")
        assert hasattr(sdk.safety.v3, "context_relevancy")
        assert hasattr(sdk.safety.v3, "pii")
        assert hasattr(sdk.safety.v3, "prompt_injection")

        # Test default version access
        assert hasattr(sdk.safety, "faithfulness")
        assert hasattr(sdk.safety, "answer_relevancy")
        assert hasattr(sdk.safety, "context_relevancy")
        assert hasattr(sdk.safety, "pii")
        assert hasattr(sdk.safety, "prompt_injection")

    def test_alignment_namespace(self, sdk: TrustwiseSDK) -> None:
        """Test alignment namespace access."""
        # Test direct version access
        assert hasattr(sdk.alignment, "v1")
        assert hasattr(sdk.alignment.v1, "clarity")
        assert hasattr(sdk.alignment.v1, "helpfulness")
        assert hasattr(sdk.alignment.v1, "formality")
        assert hasattr(sdk.alignment.v1, "simplicity")
        assert hasattr(sdk.alignment.v1, "toxicity")
        assert hasattr(sdk.alignment.v1, "sensitivity")
        assert hasattr(sdk.alignment.v1, "tone")

        # Test default version access
        assert hasattr(sdk.alignment, "clarity")
        assert hasattr(sdk.alignment, "helpfulness")
        assert hasattr(sdk.alignment, "formality")
        assert hasattr(sdk.alignment, "simplicity")
        assert hasattr(sdk.alignment, "toxicity")
        assert hasattr(sdk.alignment, "sensitivity")
        assert hasattr(sdk.alignment, "tone")

    def test_guardrails_namespace(self, sdk: TrustwiseSDK) -> None:
        """Test guardrails namespace access."""
        assert hasattr(sdk, "guardrails")
        assert callable(sdk.guardrails)

    def test_performance_namespace(self, sdk: TrustwiseSDK) -> None:
        """Test performance namespace access."""
        # Test direct version access
        assert hasattr(sdk.performance, "v1")
        assert hasattr(sdk.performance.v1, "cost")
        assert hasattr(sdk.performance.v1, "carbon")

        # Test default version access
        assert hasattr(sdk.performance, "cost")
        assert hasattr(sdk.performance, "carbon")

    def test_invalid_api_key(self) -> None:
        """Test SDK initialization with invalid API key."""
        with pytest.raises(ValueError):
            config = TrustwiseConfig(api_key="")
            TrustwiseSDK(config)

        with pytest.raises(ValueError):
            config = TrustwiseConfig(api_key=None)
            TrustwiseSDK(config)

    def test_invalid_base_url(self, api_key: str) -> None:
        """Test SDK initialization with invalid base URL."""
        with pytest.raises(ValueError):
            config = TrustwiseConfig(api_key=api_key, base_url="")
            TrustwiseSDK(config)

        with pytest.raises(ValueError):
            config = TrustwiseConfig(api_key=api_key, base_url="not-a-url")
            TrustwiseSDK(config)

    def test_namespace_version_consistency(self, sdk: TrustwiseSDK) -> None:
        """Test that namespace versions are consistent with get_versions()."""
        versions = sdk.get_versions()
        
        # Check safety version consistency
        for version in versions["safety"]:
            assert hasattr(sdk.safety, version), f"Safety namespace missing version {version}"
        
        # Check alignment version consistency
        for version in versions["alignment"]:
            assert hasattr(sdk.alignment, version), f"Alignment namespace missing version {version}"
        
        # Check performance version consistency
        for version in versions["performance"]:
            assert hasattr(sdk.performance, version), f"Performance namespace missing version {version}"

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_sdk_actual_usage(self, mock_post: Any, api_key: str) -> None:
        """Test actual SDK usage to catch import and initialization issues."""
        # Set up mock responses
        mock_post.return_value = get_mock_response("safety/v3/faithfulness")
        
        config = TrustwiseConfig(api_key=api_key)
        sdk = TrustwiseSDK(config)
        
        # Test guardrails initialization
        guardrail = sdk.guardrails(
            thresholds={
                "faithfulness": 80,
                "clarity": 70
            }
        )
        assert guardrail is not None
        
        # Test basic metric evaluation
        result = sdk.safety.faithfulness.evaluate(
            query="What is the capital of France?",
            response="The capital of France is Paris.",
            context=[{
                "node_text": "Paris is the capital of France.",
                "node_score": 0.95,
                "node_id": "doc:idx:1"
            }]
        )
        assert isinstance(result, FaithfulnessResponse)
        assert hasattr(result, "score")
        assert isinstance(result.score, float) 


import os

import pytest

from trustwise.sdk import TrustwiseSDK
from trustwise.sdk.config import TrustwiseConfig


@pytest.fixture
def api_key():
    """Get API key from environment variable or use a mock key for testing."""
    key = os.getenv("TW_API_KEY", "mock-api-key-for-testing")
    return key

@pytest.fixture
def sdk(api_key):
    """Create a Trustwise SDK instance."""
    config = TrustwiseConfig(api_key=api_key, base_url="https://random.trustwise.ai")
    return TrustwiseSDK(config)

@pytest.fixture
def sample_context():
    """Sample context data for testing."""
    return [
        {
            "node_text": "According to Crawford's Complete Auto Maintenance Guide, to check tire pressure, you'll need a tire pressure gauge. These can be purchased at auto parts stores and service stations for as little as $1-7.",
            "node_score": 0.95,
            "node_id": "doc:idx:1"
        },
        {
            "node_text": "Steps to check tire pressure: 1. Remove the valve cap from each tire 2. Align the gauge up to the valve 3. Press the gauge onto the valve with firm direct pressure and then release 4. For pen gauges, you'll see the measuring stick pushed out displaying the pressure reading 5. Fill the tire with air if needed, then recheck until you reach the desired pressure",
            "node_score": 0.92,
            "node_id": "doc:idx:2"
        },
        {
            "node_text": "The recommended tire pressure for your vehicle can be found on a sticker in the driver's door jamb, in the vehicle owner's manual, or sometimes on the inside of the fuel filler door. Most passenger vehicles have a recommended pressure between 32-35 PSI.",
            "node_score": 0.85,
            "node_id": "doc:idx:3"
        }
    ]

@pytest.fixture
def sample_query():
    """Sample query for testing."""
    return "How do I check my tire pressure?"

@pytest.fixture
def sample_response():
    """Sample response for testing."""
    return "According to Crawford's Guide, you need a tire pressure gauge which you can buy at an auto parts store or service station for about $1-7. To check pressure: remove the valve cap, press the gauge onto the valve firmly, and read the measurement. If needed, add air and recheck until you reach the right pressure." 


from typing import List, Optional

from trustwise.sdk.metrics.base import BaseMetric
from trustwise.sdk.types import (
    AnswerRelevancyRequest,
    AnswerRelevancyResponse,
    Context,
)


class AnswerRelevancyMetric:
    """Answer relevancy metric for evaluating response relevance to query."""
    
    def __init__(self, client) -> None:
        self.client = client
        self.base_url = client.config.get_safety_url("v3")
    
    def evaluate(
        self,
        *,
        query: Optional[str] = None,
        response: Optional[str] = None,
        context: Optional[Context] = None,
        **kwargs
    ) -> AnswerRelevancyResponse:
        """
        Evaluate the relevancy of a response to the query.

        Args:
            query: The query string (required)
            response: The response string (required)
            context: The context information (required)

        Returns:
            AnswerRelevancyResponse containing the evaluation results
        """
        req = BaseMetric.validate_request_model(AnswerRelevancyRequest, query=query, response=response, context=context, **kwargs)
        result = self.client._post(
            endpoint=f"{self.base_url}/answer_relevancy",
            data=req.to_dict()
        )
        return AnswerRelevancyResponse(**result)
    
    def batch_evaluate(
        self,
        inputs: List[AnswerRelevancyRequest]
    ) -> List[AnswerRelevancyResponse]:
        """Evaluate multiple inputs in a single request."""
        raise NotImplementedError("Batch evaluation not yet supported")
    
    def explain(
        self,
        *,
        query: str,
        response: str,
        context: Context,
        **kwargs
    ) -> dict:
        """Get detailed explanation of the evaluation."""
        # req = AnswerRelevancyRequest(query=query, response=response, context=context)
        raise NotImplementedError("Explanation not yet supported") 


from trustwise.sdk.metrics.alignment.v1 import AlignmentMetricsV1


class AlignmentMetrics:
    """
    Namespace for Alignment Metrics API versions.
    """

    def __init__(self, client) -> None:
        """
        Initialize the Alignment Namespace with all supported versions.
        """
        self._current_version = AlignmentMetricsV1(client)
        self.v1 = self._current_version

        # Expose all v1 metric classes directly as attributes for the preferred style
        self.clarity = self._current_version.clarity
        self.formality = self._current_version.formality
        self.helpfulness = self._current_version.helpfulness
        self.simplicity = self._current_version.simplicity
        self.tone = self._current_version.tone
        self.toxicity = self._current_version.toxicity
        self.sensitivity = self._current_version.sensitivity

    @property
    def version(self) -> str:
        """Get the current default version."""
        return self._current_version.version



from typing import List, Optional

from trustwise.sdk.types import PIIRequest, PIIResponse


class PIIMetric:
    """PII detection metric for identifying personally identifiable information."""
    
    def __init__(self, client) -> None:
        self.client = client
        self.base_url = client.config.get_safety_url("v3")
    
    def evaluate(
        self,
        *,
        text: str,
        allowlist: List[str],
        blocklist: List[str],
        **kwargs
    ) -> PIIResponse:
        """
        Evaluate the PII detection in a response.

        Args:
            text: The text to evaluate (required)
            allowlist: List of allowed PII categories (required)
            blocklist: List of blocked PII categories (required)

        Returns:
            PIIResponse containing the evaluation results
        """
        req = PIIRequest(text=text, allowlist=allowlist, blocklist=blocklist, **kwargs)
        result = self.client._post(
            endpoint=f"{self.base_url}/pii",
            data=req.model_dump() if hasattr(req, "model_dump") else req.dict()
        )
        return PIIResponse(**result)
    
    def batch_evaluate(
        self,
        texts: List[str],
        allowlist: Optional[List[str]] = None,
        blocklist: Optional[List[str]] = None
    ) -> List[PIIResponse]:
        """Evaluate multiple texts for PII content."""
        raise NotImplementedError("Batch evaluation not yet supported")
    
    def explain(
        self,
        *,
        text: str,
        allowlist: List[str],
        blocklist: List[str],
        **kwargs
    ) -> dict:
        """Get detailed explanation of the PII detection."""
        # req = PIIRequest(text=text, allowlist=allowlist, blocklist=blocklist)
        raise NotImplementedError("Explanation not yet supported") 


"""Trustwise SDK for evaluating AI-generated content."""

try:
    from importlib.metadata import version
    __version__ = version("trustwise")
except ImportError:
    # Fallback for Python < 3.8
    from pkg_resources import get_distribution
    __version__ = get_distribution("trustwise").version



from typing import Any
from unittest.mock import patch

import pytest

from trustwise.sdk import TrustwiseSDK
from trustwise.sdk.guardrails.guardrail import Guardrail
from trustwise.sdk.types import GuardrailResponse

from .helpers import get_mock_response


class TestGuardrail:
    """Test suite for Guardrail class."""

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_guardrail_initialization(self, mock_post, sdk):
        """Test guardrail initialization."""
        guardrail = Guardrail(
            trustwise_client=sdk,
            thresholds={"faithfulness": 80, "clarity": 70}
        )
        assert isinstance(guardrail, Guardrail)
        assert hasattr(guardrail, "client")
        assert guardrail.thresholds["faithfulness"] == 80
        assert guardrail.thresholds["clarity"] == 70
        assert guardrail.block_on_failure is False

    @patch("trustwise.sdk.metrics.safety.v3.FaithfulnessMetric.evaluate")
    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_evaluate(
        self,
        mock_post: Any,
        mock_faithfulness_evaluate: Any,
        sdk: TrustwiseSDK,
        sample_query: str,
        sample_context: list,
        sample_response: str
    ) -> None:
        """Test response evaluation with default metrics."""
        mock_post.return_value = get_mock_response("guardrails/evaluate")
        mock_faithfulness_evaluate.return_value = get_mock_response("safety/faithfulness")
        guardrail = Guardrail(
            trustwise_client=sdk,
            thresholds={"faithfulness": 80, "clarity": 70}
        )
        result = guardrail.evaluate(
            query=sample_query,
            context=sample_context,
            response=sample_response
        )
        
        assert isinstance(result, GuardrailResponse)
        assert isinstance(result.passed, bool)
        assert isinstance(result.blocked, bool)
        assert isinstance(result.results, dict)
        
        # Verify metric results
        results = result.results
        assert "faithfulness" in results
        assert "clarity" in results
        
        # Verify metric scores
        for metric_name, metric_result in results.items():
            assert isinstance(metric_result, dict) or hasattr(metric_result, "score")
            if isinstance(metric_result, dict) and "error" in metric_result:
                # Skip score assertion if error is present
                continue
            if hasattr(metric_result, "score"):
                assert isinstance(metric_result.score, float)
                assert 0 <= metric_result.score <= 100
            else:
                assert "score" in metric_result
                assert isinstance(metric_result["score"], float)
                assert 0 <= metric_result["score"] <= 100

    @patch("trustwise.sdk.metrics.safety.v3.FaithfulnessMetric.evaluate")
    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_evaluate_with_custom_metrics(
        self,
        mock_post: Any,
        mock_faithfulness_evaluate: Any,
        sdk: TrustwiseSDK,
        sample_query: str,
        sample_context: list,
        sample_response: str
    ) -> None:
        """Test response evaluation with custom metrics."""
        mock_post.return_value = get_mock_response("guardrails/evaluate")
        mock_faithfulness_evaluate.return_value = get_mock_response("safety/faithfulness")
        guardrail = Guardrail(
            trustwise_client=sdk,
            thresholds={"faithfulness": 80, "clarity": 70}
        )
        result = guardrail.evaluate(
            query=sample_query,
            context=sample_context,
            response=sample_response
        )
        
        assert isinstance(result, GuardrailResponse)
        assert isinstance(result.passed, bool)
        assert isinstance(result.blocked, bool)
        assert isinstance(result.results, dict)
        
        # Verify metric results
        results = result.results
        assert "faithfulness" in results
        assert "clarity" in results

    @patch("trustwise.sdk.metrics.safety.v3.FaithfulnessMetric.evaluate")
    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_evaluate_without_blocking(
        self,
        mock_post: Any,
        mock_faithfulness_evaluate: Any,
        sdk: TrustwiseSDK,
        sample_query: str,
        sample_context: list,
        sample_response: str
    ) -> None:
        """Test response evaluation without blocking."""
        mock_post.return_value = get_mock_response("guardrails/evaluate")
        mock_faithfulness_evaluate.return_value = get_mock_response("safety/faithfulness")
        guardrail = Guardrail(
            trustwise_client=sdk,
            thresholds={"faithfulness": 80, "clarity": 70},
            block_on_failure=False
        )
        result = guardrail.evaluate(
            query=sample_query,
            context=sample_context,
            response=sample_response
        )
        
        assert isinstance(result, GuardrailResponse)
        assert isinstance(result.passed, bool)
        assert isinstance(result.blocked, bool)
        assert result.blocked is False  # Should not block even if failed

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_invalid_thresholds(self, mock_post, sdk):
        """Test handling of invalid thresholds."""
        with pytest.raises(ValueError):
            Guardrail(
                trustwise_client=sdk,
                thresholds={
                    "faithfulness": 150  # Invalid threshold > 100
                }
            )
        
        with pytest.raises(ValueError):
            Guardrail(
                trustwise_client=sdk,
                thresholds={
                    "prompt_injection": 150  # Invalid threshold > 100
                }
            )

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_missing_parameters(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK,
        sample_query: str,
        sample_context: list
    ) -> None:
        """Test handling of missing parameters."""
        mock_post.side_effect = ValueError("Missing parameters")
        guardrail = Guardrail(
            trustwise_client=sdk,
            thresholds={"faithfulness": 80}
        )
        with pytest.raises(ValueError):
            guardrail.evaluate(
                query=sample_query,
                context=sample_context,
                response=None
            )

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_evaluate_with_prompt_injection(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK,
        sample_query: str,
        sample_context: list,
        sample_response: str
    ) -> None:
        """Test response evaluation with prompt injection detection."""
        mock_post.return_value = get_mock_response("guardrails/evaluate")
        guardrail = Guardrail(
            trustwise_client=sdk,
            thresholds={"prompt_injection": 0.1}  # Low threshold to detect injection attempts
        )
        result = guardrail.evaluate(
            query=sample_query,
            context=sample_context,
            response=sample_response
        )
        
        assert isinstance(result, GuardrailResponse)
        assert isinstance(result.passed, bool)
        assert isinstance(result.blocked, bool)
        assert isinstance(result.results, dict)
        
        # Verify prompt injection metric
        results = result.results
        assert "prompt_injection" in results
        assert isinstance(results["prompt_injection"], dict)
        assert "result" in results["prompt_injection"]
        assert "passed" in results["prompt_injection"]
        assert "result" in results["prompt_injection"]
        assert isinstance(results["prompt_injection"]["result"], dict)
        assert "score" in results["prompt_injection"]["result"]
        assert isinstance(results["prompt_injection"]["result"]["score"], float)
        assert 0 <= results["prompt_injection"]["result"]["score"] <= 100 


"""Simple configuration for Trustwise API with environment variable support."""

import os
from urllib.parse import urljoin, urlparse

from pydantic import BaseModel, Field, ValidationError


class TrustwiseConfig(BaseModel):
    """Configuration for Trustwise API.
    
    This class handles configuration for the Trustwise API, including authentication and endpoint URLs.
    It supports both direct initialization and environment variables.
    
    Args:
        api_key (str, Optional): API key for authentication. If not provided, will be read from
            TW_API_KEY environment variable. Required for API access.
        base_url (str, Optional): Base URL for API endpoints. If not provided, will be read from
            TW_BASE_URL environment variable or default to https://api.trustwise.ai.
    
    Environment Variables:
        - TW_API_KEY: API key for authentication
        - TW_BASE_URL: Base URL for API endpoints (defaults to https://api.trustwise.ai)
    
    Example:
        >>> from trustwise.sdk.config import TrustwiseConfig
        >>> # Using environment variables
        >>> config = TrustwiseConfig()
        >>> # Using direct initialization
        >>> config = TrustwiseConfig(api_key="your-api-key", base_url="https://api.trustwise.ai")
    
    Raises:
        ValueError: If API key is missing or invalid, or if base URL is invalid
    """
    
    api_key: str = Field(
        default=None,
        description="API key for authentication"
    )
    
    base_url: str = Field(
        default_factory=lambda: os.getenv("TW_BASE_URL") or "https://api.trustwise.ai",
        description="Base URL for API endpoints"
    )
    
    def __init__(self, **data) -> None:
        # Check for API key in environment if not provided
        if "api_key" not in data:
            data["api_key"] = os.getenv("TW_API_KEY")
            
        # Validate API key before Pydantic initialization
        if not data.get("api_key"):
            raise ValueError("API key must be provided either directly or through TW_API_KEY environment variable")
            
        try:
            super().__init__(**data)
        except ValidationError as e:
            # Convert Pydantic validation errors to our own ValueError
            if "api_key" in str(e):
                raise ValueError("API key must be a valid string") from e
            if "base_url" in str(e):
                raise ValueError("Base URL must be a valid string") from e
            raise ValueError(f"Invalid configuration: {e!s}") from e
            
        self._validate_url()
        # Ensure base_url ends with a slash for consistent path joining
        if not self.base_url.endswith("/"):
            self.base_url = f"{self.base_url}/"
    
    def _validate_url(self) -> None:
        """Validate base URL format."""
        try:
            parsed = urlparse(self.base_url)
            if not all([parsed.scheme, parsed.netloc]):
                raise ValueError("Invalid base URL format")
        except (ValueError, AttributeError) as e:
            raise ValueError(f"Invalid base URL: {e!s}") from e
    
    def get_safety_url(self, version: str) -> str:
        """Get safety API URL for specified version."""
        return urljoin(self.base_url, f"safety/{version}")
    
    def get_alignment_url(self, version: str) -> str:
        """Get alignment API URL for specified version."""
        return urljoin(self.base_url, f"alignment/{version}")

    def get_performance_url(self, version: str) -> str:
        """Get performance API URL for specified version."""
        return urljoin(self.base_url, f"performance/{version}") 


from typing import Any, Dict, List
from unittest.mock import patch

import pytest
from pydantic import ValidationError

from trustwise.sdk import TrustwiseSDK
from trustwise.sdk.exceptions import TrustwiseValidationError
from trustwise.sdk.types import PromptInjectionRequest

from .helpers import get_mock_response


class TestSafetyMetricsV3:
    """Test suite for Safety Metrics API v3."""

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_faithfulness(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK,
        sample_query: str,
        sample_context: List[Dict[str, Any]],
        sample_response: str
    ) -> None:
        """Test faithfulness evaluation."""
        mock_post.return_value = get_mock_response("safety/v3/faithfulness")
        result = sdk.safety.v3.faithfulness.evaluate(
            query=sample_query,
            context=sample_context,
            response=sample_response
        )
        from trustwise.sdk.types import FaithfulnessResponse
        assert isinstance(result, FaithfulnessResponse)
        assert isinstance(result.score, float)
        assert 0 <= result.score <= 100
        assert isinstance(result.facts, list)
        for fact in result.facts:
            assert hasattr(fact, "statement")
            assert hasattr(fact, "label")
            assert hasattr(fact, "prob")
            assert hasattr(fact, "sentence_span")

    def test_faithfulness_missing_context(
        self,
        sdk: TrustwiseSDK,
        sample_query: str,
        sample_response: str
    ) -> None:
        """Test faithfulness evaluation with missing context."""
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.faithfulness.evaluate(
                query=sample_query,
                response=sample_response,
                context=None
            )
        assert "Error in 'FaithfulnessRequest'" in str(excinfo.value)
        assert "'context'" in str(excinfo.value)

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_answer_relevancy(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK,
        sample_query: str,
        sample_context: List[Dict[str, Any]],
        sample_response: str
    ) -> None:
        """Test answer relevancy evaluation."""
        mock_post.return_value = get_mock_response("safety/v3/answer_relevancy")
        result = sdk.safety.v3.answer_relevancy.evaluate(
            query=sample_query,
            context=sample_context,
            response=sample_response
        )
        
        from trustwise.sdk.types import AnswerRelevancyResponse
        assert isinstance(result, AnswerRelevancyResponse)
        assert isinstance(result.score, float)
        assert 0 <= result.score <= 100  # Scores are percentages
        if hasattr(result, "generated_question"):
            assert isinstance(result.generated_question, str)

    def test_answer_relevancy_missing_context(
        self,
        sdk: TrustwiseSDK,
        sample_query: str,
        sample_response: str
    ) -> None:
        """Test answer relevancy evaluation with missing context."""
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.answer_relevancy.evaluate(
                query=sample_query,
                response=sample_response,
                context=None
            )
        assert "Error in 'AnswerRelevancyRequest'" in str(excinfo.value)
        assert "'context'" in str(excinfo.value)

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_context_relevancy(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK,
        sample_query: str,
        sample_context: List[Dict[str, Any]],
        sample_response: str
    ) -> None:
        """Test context relevancy evaluation."""
        mock_post.return_value = get_mock_response("safety/v3/context_relevancy")
        result = sdk.safety.v3.context_relevancy.evaluate(
            query=sample_query,
            context=sample_context,
            response=sample_response
        )
        from trustwise.sdk.types import ContextRelevancyResponse
        assert isinstance(result, ContextRelevancyResponse)
        assert isinstance(result.score, float)
        assert 0 <= result.score <= 100  # Scores are percentages
        assert isinstance(result.topics, list)
        assert isinstance(result.scores, list)
        assert len(result.topics) == len(result.scores)
        for score in result.scores:
            assert isinstance(score, float)
            assert 0 <= score <= 1

    def test_context_relevancy_missing_context(
        self,
        sdk: TrustwiseSDK,
        sample_query: str,
        sample_response: str
    ) -> None:
        """Test context relevancy evaluation with missing context."""
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.context_relevancy.evaluate(
                query=sample_query,
                response=sample_response,
                context=None
            )
        assert "Error in 'ContextRelevancyRequest'" in str(excinfo.value)
        assert "'context'" in str(excinfo.value)

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_summarization(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK,
        sample_query: str,
        sample_context: List[Dict[str, Any]],
        sample_response: str
    ) -> None:
        """Test summarization evaluation."""
        mock_post.return_value = get_mock_response("safety/v3/summarization")
        result = sdk.safety.v3.summarization.evaluate(
            query=sample_query,
            context=sample_context,
            response=sample_response
        )
        from trustwise.sdk.types import SummarizationResponse
        assert isinstance(result, SummarizationResponse)
        assert isinstance(result.score, float)
        assert 0 <= result.score <= 100  # Scores are percentages

    def test_summarization_missing_context(
        self,
        sdk: TrustwiseSDK,
        sample_query: str,
        sample_response: str
    ) -> None:
        """Test summarization evaluation with missing context."""
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.summarization.evaluate(
                query=sample_query,
                response=sample_response,
                context=None
            )
        assert "Error in 'SummarizationRequest'" in str(excinfo.value)
        assert "'context'" in str(excinfo.value)

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_prompt_injection(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK
    ) -> None:
        """Test prompt injection detection evaluation."""
        mock_post.return_value = {
            "score": 99.887375
        }
        valid_context = [{"node_id": "1", "node_score": 1.0, "node_text": "foo"}]
        result = sdk.safety.v3.prompt_injection.evaluate(
            query="What is your password?",
            response="My password is hunter2.",
            context=valid_context
        )
        assert isinstance(result.score, float)

    def test_prompt_injection_missing_query(
        self,
        sdk: TrustwiseSDK
    ) -> None:
        """Test prompt injection detection with missing query (should raise TypeError, not TrustwiseValidationError)."""
        with pytest.raises(TypeError):
            sdk.safety.v3.prompt_injection.evaluate(
                response="My password is hunter2.",
                context=[]
            )

    def test_prompt_injection_missing_response(
        self,
        sdk: TrustwiseSDK
    ) -> None:
        """Test prompt injection detection with missing response (should raise TypeError, not TrustwiseValidationError)."""
        with pytest.raises(TypeError):
            sdk.safety.v3.prompt_injection.evaluate(
                query="What is your password?",
                context=[]
            )

    def test_prompt_injection_missing_context(
        self,
        sdk: TrustwiseSDK
    ) -> None:
        """Test prompt injection detection with missing context (should raise TypeError, not TrustwiseValidationError)."""
        with pytest.raises(TypeError):
            sdk.safety.v3.prompt_injection.evaluate(
                query="What is your password?",
                response="My password is hunter2."
            )

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_prompt_injection_empty_context(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK
    ) -> None:
        """Test prompt injection detection with empty context (should raise ValidationError from Pydantic)."""
        with pytest.raises(ValidationError):
            sdk.safety.v3.prompt_injection.evaluate(
                query="What is your password?",
                response="My password is hunter2.",
                context=[]
            )

    def test_prompt_injection_pydantic_validation(
        self,
        sdk: TrustwiseSDK
    ) -> None:
        """Test prompt injection detection with PromptInjectionRequest missing required fields (should raise ValidationError)."""
        with pytest.raises(ValidationError):
            # Missing context
            PromptInjectionRequest(query="foo", response="bar")
        with pytest.raises(ValidationError):
            # Missing query
            PromptInjectionRequest(response="bar", context=[{"node_id": "1", "node_score": 1.0, "node_text": "foo"}])
        with pytest.raises(ValidationError):
            # Missing response
            PromptInjectionRequest(query="foo", context=[{"node_id": "1", "node_score": 1.0, "node_text": "foo"}])

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_invalid_input(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK
    ) -> None:
        """Test handling of invalid input."""
        mock_post.side_effect = ValueError("Invalid input")
        with pytest.raises(ValueError):
            sdk.safety.v3.faithfulness.evaluate(
                query="",
                context=[],
                response=""
            )

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_missing_parameters(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK
    ) -> None:
        """Test handling of missing parameters (should raise TrustwiseValidationError for present-but-invalid, e.g., None)."""
        mock_post.side_effect = ValueError("Missing parameters")
        with pytest.raises(TrustwiseValidationError):
            sdk.safety.v3.faithfulness.evaluate(
                query="test",
                context=None,
                response="test"
            )

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_pii(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK
    ) -> None:
        """Test PII detection evaluation."""
        mock_post.return_value = {
            "identified_pii": [
                {
                    "interval": [0, 5],
                    "string": "Hello",
                    "category": "blocklist"
                }
            ]
        }
        result = sdk.safety.v3.pii.evaluate(
            text="Contact me at john@example.com.",
            allowlist=["EMAIL"],
            blocklist=["PHONE"]
        )
        from trustwise.sdk.types import PIIEntity, PIIResponse
        assert isinstance(result, PIIResponse)
        assert isinstance(result.identified_pii, list)
        pii = result.identified_pii[0]
        assert isinstance(pii, PIIEntity)
        assert pii.interval == [0, 5]
        assert pii.string == "Hello"
        assert pii.category == "blocklist"

    def test_pii_missing_text(
        self,
        sdk: TrustwiseSDK
    ) -> None:
        """Test PII detection with missing fields."""
        with pytest.raises(TypeError):
            sdk.safety.v3.pii.evaluate()

    def test_faithfulness_pydantic_validation(self):
        """Test FaithfulnessRequest missing required fields (should raise ValidationError)."""
        from pydantic import ValidationError

        from trustwise.sdk.types import FaithfulnessRequest
        # Missing context
        with pytest.raises(ValidationError):
            FaithfulnessRequest(query="foo", response="bar")
        # Missing query
        with pytest.raises(ValidationError):
            FaithfulnessRequest(response="bar", context=[{"node_id": "1", "node_score": 1.0, "node_text": "foo"}])
        # Missing response
        with pytest.raises(ValidationError):
            FaithfulnessRequest(query="foo", context=[{"node_id": "1", "node_score": 1.0, "node_text": "foo"}]) 

    def test_faithfulness_missing_required_fields(self, sdk: TrustwiseSDK, sample_query: str, sample_response: str):
        """Test faithfulness evaluation with missing required fields."""
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.faithfulness.evaluate(response=sample_response, context=[])
        assert "'query'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.faithfulness.evaluate(query=sample_query, context=[])
        assert "'response'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.faithfulness.evaluate(query=sample_query, response=sample_response)
        assert "'context'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.faithfulness.evaluate()
        assert "'query'" in str(excinfo.value)
        assert "'response'" in str(excinfo.value)
        assert "'context'" in str(excinfo.value)

    def test_answer_relevancy_missing_required_fields(self, sdk: TrustwiseSDK, sample_query: str, sample_response: str):
        """Test answer relevancy evaluation with missing required fields."""
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.answer_relevancy.evaluate(response=sample_response, context=[])
        assert "'query'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.answer_relevancy.evaluate(query=sample_query, context=[])
        assert "'response'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.answer_relevancy.evaluate(query=sample_query, response=sample_response)
        assert "'context'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.answer_relevancy.evaluate()
        assert "'query'" in str(excinfo.value)
        assert "'response'" in str(excinfo.value)
        assert "'context'" in str(excinfo.value)

    def test_context_relevancy_missing_required_fields(self, sdk: TrustwiseSDK, sample_query: str, sample_response: str):
        """Test context relevancy evaluation with missing required fields."""
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.context_relevancy.evaluate(response=sample_response, context=[])
        assert "'query'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.context_relevancy.evaluate(query=sample_query, context=[])
        assert "'response'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.context_relevancy.evaluate(query=sample_query, response=sample_response)
        assert "'context'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.context_relevancy.evaluate()
        assert "'query'" in str(excinfo.value)
        assert "'response'" in str(excinfo.value)
        assert "'context'" in str(excinfo.value)

    def test_summarization_missing_required_fields(self, sdk: TrustwiseSDK, sample_query: str, sample_response: str):
        """Test summarization evaluation with missing required fields."""
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.summarization.evaluate(response=sample_response, context=[])
        assert "'query'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.summarization.evaluate(query=sample_query, context=[])
        assert "'response'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.summarization.evaluate(query=sample_query, response=sample_response)
        assert "'context'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.safety.v3.summarization.evaluate()
        assert "'query'" in str(excinfo.value)
        assert "'response'" in str(excinfo.value)
        assert "'context'" in str(excinfo.value)

    def test_prompt_injection_missing_required_fields(self, sdk: TrustwiseSDK):
        """Test prompt injection detection with missing required fields (should raise TypeError for omitted args)."""
        with pytest.raises(TypeError):
            sdk.safety.v3.prompt_injection.evaluate(response="foo", context=[])
        with pytest.raises(TypeError):
            sdk.safety.v3.prompt_injection.evaluate(query="foo", context=[])
        with pytest.raises(TypeError):
            sdk.safety.v3.prompt_injection.evaluate(query="foo", response="foo")
        with pytest.raises(TypeError):
            sdk.safety.v3.prompt_injection.evaluate()

    def test_pii_missing_required_fields(self, sdk: TrustwiseSDK):
        """Test PII detection with missing required fields (should raise TypeError for omitted args)."""
        with pytest.raises(TypeError):
            sdk.safety.v3.pii.evaluate(allowlist=["EMAIL"], blocklist=["PHONE"])
        with pytest.raises(TypeError):
            sdk.safety.v3.pii.evaluate(text="foo", blocklist=["PHONE"])
        with pytest.raises(TypeError):
            sdk.safety.v3.pii.evaluate(text="foo", allowlist=["EMAIL"])
        with pytest.raises(TypeError):
            sdk.safety.v3.pii.evaluate() 



# Import required packages
import os
from dotenv import load_dotenv
from trustwise.sdk import TrustwiseSDK
from trustwise.sdk.config import TrustwiseConfig

# Load environment variables from .env file
load_dotenv()

config = TrustwiseConfig()  # Automatically uses TW_API_KEY from environment
trustwise = TrustwiseSDK(config)

# # Try to evaluate with invalid API key
# invalid_config = TrustwiseConfig(api_key="invalid_key")
# invalid_sdk = TrustwiseSDK(invalid_config)
# result = invalid_sdk.safety.v3.faithfulness.evaluate(
#     query="What is the capital of France?",
#     response="The capital of France is Paris.",
#     context=[]
# )

context = [{
    "node_text": "Paris is the capital of France. It is known for the Eiffel Tower and the Louvre Museum.",
    "node_score": 0.95,
    "node_id": "doc:idx:1"
}]
# # Try to evaluate with invalid input (missing required field)
# result = trustwise.safety.v3.faithfulness.evaluate(
#     query="What is the capital of France?",
#     # Missing 'response' parameter
#     context=context
# )


result = trustwise.safety.v3.faithfulness.evaluate(
    query="What is the capital of France?",
    response="",
    context=context
)


from typing import Any
from unittest.mock import patch

import pytest

from trustwise.sdk import TrustwiseSDK
from trustwise.sdk.exceptions import TrustwiseValidationError
from trustwise.sdk.types import (
    ClarityResponse,
    FormalityResponse,
    HelpfulnessResponse,
    SensitivityResponse,
    SimplicityResponse,
    ToneResponse,
    ToxicityResponse,
)

from .helpers import get_mock_response


class TestAlignmentMetricsV1:
    """Test suite for Alignment Metrics API v1."""

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_clarity(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK,
        sample_query: str,
        sample_response: str
    ) -> None:
        """Test clarity evaluation."""
        mock_post.return_value = get_mock_response("alignment/v1/clarity")
        result = sdk.alignment.v1.clarity.evaluate(
            query=sample_query,
            response=sample_response
        )
        
        assert isinstance(result, ClarityResponse)
        assert isinstance(result.score, float)
        assert 0 <= result.score <= 100  # Scores are percentages
        # Test JSON output
        json_str = result.to_json()
        import json as _json
        data = _json.loads(json_str)
        assert data["score"] == result.score

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_helpfulness(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK,
        sample_query: str,
        sample_response: str
    ) -> None:
        """Test helpfulness evaluation."""
        mock_post.return_value = get_mock_response("alignment/v1/helpfulness")
        result = sdk.alignment.v1.helpfulness.evaluate(
            query=sample_query,
            response=sample_response
        )
        assert isinstance(result, HelpfulnessResponse)
        assert isinstance(result.score, float)
        assert 0 <= result.score <= 100  # Scores are percentages
        # Test JSON output
        json_str = result.to_json()
        import json as _json
        data = _json.loads(json_str)
        assert data["score"] == result.score

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_formality(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK,
        sample_response: str
    ) -> None:
        """Test formality evaluation."""
        mock_post.return_value = get_mock_response("alignment/v1/formality")
        result = sdk.alignment.v1.formality.evaluate(
            response=sample_response
        )
        assert isinstance(result, FormalityResponse)
        assert isinstance(result.score, float)
        assert isinstance(result.sentences, list)
        assert isinstance(result.scores, list)
        assert 0 <= result.score <= 100  # Scores are percentages
        # Test JSON output
        json_str = result.to_json()
        import json as _json
        data = _json.loads(json_str)
        assert data["score"] == result.score
        assert data["sentences"] == result.sentences
        assert data["scores"] == result.scores

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_simplicity(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK,
        sample_response: str
    ) -> None:
        """Test simplicity evaluation."""
        mock_post.return_value = get_mock_response("alignment/v1/simplicity")
        result = sdk.alignment.v1.simplicity.evaluate(
            response=sample_response
        )
        assert isinstance(result, SimplicityResponse)
        assert isinstance(result.score, float)
        assert 0 <= result.score <= 100  # Scores are percentages
        # Test JSON output
        json_str = result.to_json()
        import json as _json
        data = _json.loads(json_str)
        assert data["score"] == result.score

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_sensitivity(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK,
        sample_query: str,
        sample_response: str
    ) -> None:
        """Test sensitivity evaluation."""
        mock_post.return_value = get_mock_response("alignment/v1/sensitivity")
        result = sdk.alignment.v1.sensitivity.evaluate(
            response=sample_response,
            topics=["health", "finance"],
            query=sample_query
        )
        assert isinstance(result, SensitivityResponse)
        assert isinstance(result.scores, dict)
        for topic, score in result.scores.items():
            assert isinstance(score, float)
            assert 0 <= score <= 100  # Scores are percentages
        # Test JSON output
        json_str = result.to_json()
        import json as _json
        data = _json.loads(json_str)
        assert data["scores"] == result.scores

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_toxicity(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK,
        sample_query: str,
        sample_response: str
    ) -> None:
        """Test toxicity evaluation."""
        mock_post.return_value = get_mock_response("alignment/v1/toxicity")
        result = sdk.alignment.v1.toxicity.evaluate(
            query=sample_query,
            response=sample_response
        )
        assert isinstance(result, ToxicityResponse)
        assert isinstance(result.labels, list)
        assert isinstance(result.scores, list)
        assert len(result.labels) == len(result.scores)
        for score in result.scores:
            assert isinstance(score, float)
            assert 0 <= score <= 1  # Toxicity scores are between 0 and 1
        # Test JSON output
        json_str = result.to_json()
        import json as _json
        data = _json.loads(json_str)
        assert data["labels"] == result.labels
        assert data["scores"] == result.scores

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_tone(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK,
        sample_response: str
    ) -> None:
        """Test tone evaluation."""
        mock_post.return_value = get_mock_response("alignment/v1/tone")
        result = sdk.alignment.v1.tone.evaluate(
            response=sample_response
        )
        assert isinstance(result, ToneResponse)
        assert isinstance(result.labels, list)
        assert isinstance(result.scores, list)
        assert len(result.labels) == len(result.scores)
        for score in result.scores:
            assert isinstance(score, float)
            assert 0 <= score <= 100  # Scores are percentages
        # Test JSON output
        json_str = result.to_json()
        import json as _json
        data = _json.loads(json_str)
        assert data["labels"] == result.labels
        assert data["scores"] == result.scores

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_invalid_input(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK
    ) -> None:
        """Test handling of invalid input."""
        mock_post.side_effect = ValueError("Invalid input")
        with pytest.raises(ValueError):
            sdk.alignment.v1.clarity.evaluate(
                query="",
                response=""
            )

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_missing_parameters(
        self,
        mock_post: Any,
        sdk: TrustwiseSDK
    ) -> None:
        """Test handling of missing parameters (should raise TrustwiseValidationError for present-but-invalid, e.g., None)."""
        mock_post.side_effect = ValueError("Missing parameters")
        with pytest.raises(TrustwiseValidationError):
            sdk.alignment.v1.clarity.evaluate(
                query="test",
                response=None
            )

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_clarity_missing_required_fields(self, mock_post: Any, sdk: TrustwiseSDK):
        """Test clarity evaluation with missing required fields."""
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.alignment.v1.clarity.evaluate(response="foo")
        assert "'query'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.alignment.v1.clarity.evaluate(query="foo")
        assert "'response'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.alignment.v1.clarity.evaluate()
        assert "'query'" in str(excinfo.value)
        assert "'response'" in str(excinfo.value)

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_helpfulness_missing_required_fields(self, mock_post: Any, sdk: TrustwiseSDK):
        """Test helpfulness evaluation with missing required fields."""
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.alignment.v1.helpfulness.evaluate(response="foo")
        assert "'query'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.alignment.v1.helpfulness.evaluate(query="foo")
        assert "'response'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.alignment.v1.helpfulness.evaluate()
        assert "'query'" in str(excinfo.value)
        assert "'response'" in str(excinfo.value)

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_formality_missing_required_fields(self, mock_post: Any, sdk: TrustwiseSDK):
        """Test formality evaluation with missing required fields."""
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.alignment.v1.formality.evaluate()
        assert "'response'" in str(excinfo.value)

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_simplicity_missing_required_fields(self, mock_post: Any, sdk: TrustwiseSDK):
        """Test simplicity evaluation with missing required fields."""
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.alignment.v1.simplicity.evaluate()
        assert "'response'" in str(excinfo.value)

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_sensitivity_missing_required_fields(self, mock_post: Any, sdk: TrustwiseSDK):
        """Test sensitivity evaluation with missing required fields."""
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.alignment.v1.sensitivity.evaluate(topics=["foo"])
        assert "'response'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.alignment.v1.sensitivity.evaluate(response="foo")
        assert "'topics'" in str(excinfo.value)
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.alignment.v1.sensitivity.evaluate()
        assert "'response'" in str(excinfo.value)
        assert "'topics'" in str(excinfo.value)

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_toxicity_missing_required_fields(self, mock_post: Any, sdk: TrustwiseSDK):
        """Test toxicity evaluation with missing required fields (should raise TypeError for omitted args)."""
        with pytest.raises(TypeError):
            sdk.alignment.v1.toxicity.evaluate(response="foo")
        with pytest.raises(TypeError):
            sdk.alignment.v1.toxicity.evaluate(query="foo")
        with pytest.raises(TypeError):
            sdk.alignment.v1.toxicity.evaluate()

    @patch("trustwise.sdk.client.TrustwiseClient._post")
    def test_tone_missing_required_fields(self, mock_post: Any, sdk: TrustwiseSDK):
        """Test tone evaluation with missing required fields."""
        with pytest.raises(TrustwiseValidationError) as excinfo:
            sdk.alignment.v1.tone.evaluate()
        assert "'response'" in str(excinfo.value) 


from typing import Dict, List, Literal, Optional

from pydantic import (
    BaseModel,
    Field,
    PositiveFloat,
    PositiveInt,
    constr,
    field_validator,
)


class SDKBaseModel(BaseModel):
    class Config:
        extra = "forbid"

    def to_json(self, **kwargs) -> str:
        """
        Return a JSON string representation of the model.
        Ensures valid JSON output regardless of Pydantic version.
        Always excludes None fields by default.
        """
        kwargs.setdefault("exclude_none", True)
        return self.model_dump_json(**kwargs)

    def to_dict(self, **kwargs) -> dict:
        """
        Return a dict representation of the model, always excluding None fields by default.
        """
        kwargs.setdefault("exclude_none", True)
        return super().model_dump(**kwargs)

    @classmethod
    def validate_score_range(cls, v, min_value: float, max_value: float, label: str) -> object:
        if isinstance(v, float) or isinstance(v, int):
            if not (min_value <= v <= max_value):
                raise ValueError(f"{label} score {v} must be between {min_value} and {max_value}")
        elif isinstance(v, list):
            for s in v:
                if not (min_value <= s <= max_value):
                    raise ValueError(f"{label} score {s} must be between {min_value} and {max_value}")
        elif isinstance(v, dict):
            for k, s in v.items():
                if not (min_value <= s <= max_value):
                    raise ValueError(f"{label} score for '{k}' was {s}, must be between {min_value} and {max_value}")
        return v

    @staticmethod
    def format_validation_error(model_cls, validation_error) -> str:
        """
        Format a Pydantic ValidationError into a user-friendly error message using field types and descriptions.
        Distinguishes between missing and invalid arguments, and sets the error prefix accordingly.
        """
        errors = validation_error.errors()
        model_fields = getattr(model_cls, "model_fields", getattr(model_cls, "__fields__", {}))
        messages = []
        error_types = set()
        def get_type_str(field_type) -> str:
            origin = getattr(field_type, "__origin__", None)
            args = getattr(field_type, "__args__", None)
            if origin and args:
                origin_name = getattr(origin, "__name__", str(origin))
                args_str = ", ".join(get_type_str(a) for a in args)
                return f"{origin_name}[{args_str}]"
            return getattr(field_type, "__name__", str(field_type))
        for err in errors:
            loc = err.get("loc", [])
            field = loc[0] if loc else None
            actual_value = err.get("input", None)
            actual_type = type(actual_value).__name__ if actual_value is not None else "NoneType"
            err_type = err.get("type", "")
            if err_type.startswith("missing"):
                error_types.add("missing")
            else:
                error_types.add("invalid")
            if field and field in model_fields:
                field_info = model_fields[field]
                field_type = getattr(field_info, "annotation", getattr(field_info, "type_", None))
                type_str = get_type_str(field_type) if field_type else "unknown"
                if err_type.startswith("missing"):
                    messages.append(f"'{field}' (missing required argument, expected type: {type_str})")
                else:
                    messages.append(
                        f"'{field}' (invalid value: expected type: {type_str}, got: {actual_type} [value: {actual_value!r}])"
                    )
            elif field:
                if err_type.startswith("missing"):
                    messages.append(f"'{field}' (missing required argument)")
                else:
                    messages.append(f"'{field}' (invalid value: got: {actual_type} [value: {actual_value!r}])")
            else:
                messages.append(str(err))
        model_name = getattr(model_cls, "__name__", str(model_cls))
        if error_types == {"missing"}:
            prefix = "Missing required arguments"
        elif error_types == {"invalid"}:
            prefix = "Invalid arguments"
        else:
            prefix = "Invalid or missing arguments"
        return f"Error in '{model_name}': {prefix}: {', '.join(messages)}. Refer to the documentation: https://trustwiseai.github.io/tw-docs/docs/intro"

class Fact(SDKBaseModel):
    """
    A fact extracted from a response with its verification status.
    :param statement: The fact statement text.
    :param label: The label for the fact.
    :param prob: The probability/confidence for the fact.
    :param sentence_span: The character span of the statement in the response.
    """
    statement: str
    label: str
    prob: float
    sentence_span: List[int]

Facts = List[Fact]

class FaithfulnessResponse(SDKBaseModel):
    """
    Response type for faithfulness evaluation.
    :param score: The faithfulness score.
    :param facts: List of facts extracted from the response.
    """
    score: float = Field(..., ge=0, le=100, description="Faithfulness score (0-100)")
    facts: Facts
    @field_validator("score", mode="before")
    @classmethod
    def check_score(cls, v: float) -> float:
        return cls.validate_score_range(v, 0, 100, "Faithfulness")

class AnswerRelevancyResponse(SDKBaseModel):
    """
    Response type for answer relevancy evaluation.

    :param score: The answer relevancy score.
    :param generated_question: The generated question for evaluation.
    """
    score: float = Field(..., ge=0, le=100, description="Answer relevancy score (0-100)")
    generated_question: str
    @field_validator("score", mode="before")
    @classmethod
    def check_score(cls, v: float) -> float:
        return cls.validate_score_range(v, 0, 100, "Answer relevancy")

class ContextNode(SDKBaseModel):
    """
    A single context node with its metadata.

    :param node_id: The unique identifier for the context node.
    :param node_score: The score associated with the node.
    :param node_text: The text content of the node.
    """
    node_id: str
    node_score: float
    node_text: str

# Define Context as a list of ContextNode (Pydantic model)
Context = List[ContextNode]
"""A list of ContextNode objects representing the context for evaluation."""

class ContextRelevancyRequest(SDKBaseModel):
    """
    Request type for context relevancy evaluation.
    :param query: The input query string.
    :param context: The context information (required).
    :param response: The response to evaluate.
    """
    query: str
    context: Context
    response: str

class ContextRelevancyResponse(SDKBaseModel):
    """
    Response type for context relevancy evaluation.
    :param score: The context relevancy score.
    :param topics: List of topics identified.
    :param scores: List of scores for each topic.
    """
    score: float = Field(..., ge=0, le=100, description="Context relevancy score (0-100)")
    topics: List[str]
    scores: List[float]
    @field_validator("score", mode="before")
    @classmethod
    def check_score(cls, v: float) -> float:
        return cls.validate_score_range(v, 0, 100, "Context relevancy")
    @field_validator("scores", mode="before")
    @classmethod
    def check_scores(cls, v: List[float]) -> List[float]:
        return cls.validate_score_range(v, 0, 100, "Context relevancy")

class SummarizationRequest(SDKBaseModel):
    """
    Request type for summarization evaluation.
    :param query: The input query string.
    :param response: The response to evaluate.
    :param context: The context information (required).
    """
    query: str
    response: str
    context: Context

class SummarizationResponse(SDKBaseModel):
    """
    Response type for summarization quality evaluation.
    :param score: The summarization score.
    """
    score: float = Field(..., ge=0, le=100, description="Summarization score (0-100)")
    @field_validator("score", mode="before")
    @classmethod
    def check_score(cls, v: float) -> float:
        return cls.validate_score_range(v, 0, 100, "Summarization")

class PIIEntity(SDKBaseModel):
    """
    A detected piece of personally identifiable information.
    :param interval: The [start, end] indices of the PII in the text.
    :param string: The detected PII string.
    :param category: The PII category.
    """
    interval: List[int]
    string: str
    category: str

class PIIRequest(SDKBaseModel):
    """
    Request type for PII detection.
    :param text: The text to evaluate.
    :param allowlist: List of allowed PII categories.
    :param blocklist: List of blocked PII categories.
    """
    text: str
    allowlist: List[str]
    blocklist: List[str]

class PIIResponse(SDKBaseModel):
    """
    Response type for PII detection.
    :param identified_pii: List of detected PII entities.
    """
    identified_pii: List[PIIEntity]

class PromptInjectionRequest(SDKBaseModel):
    """
    Request type for prompt injection detection.
    :param query: The input query string.
    :param response: The response to evaluate.
    :param context: The context information (required).
    """
    query: str
    response: str
    context: Context = Field(..., min_items=1, description="A non-empty list of ContextNode objects.")

class PromptInjectionResponse(SDKBaseModel):
    """
    Response type for prompt injection detection.
    :param score: The prompt injection score.
    """
    score: float = Field(..., ge=0, le=100, description="Prompt injection score (0-100)")
    @field_validator("score", mode="before")
    @classmethod
    def check_score(cls, v: float) -> float:
        return cls.validate_score_range(v, 0, 100, "Prompt injection")

class ClarityRequest(SDKBaseModel):
    """
    Request type for clarity evaluation.
    :param query: The input query string.
    :param response: The response to evaluate.
    """
    query: str
    response: str

class ClarityResponse(SDKBaseModel):
    """
    Response type for clarity evaluation.
    :param score: The overall clarity score.
    """
    score: float = Field(..., ge=0, le=100, description="Clarity score (0-100)")
    @field_validator("score", mode="before")
    @classmethod
    def check_score(cls, v: float) -> float:
        return cls.validate_score_range(v, 0, 100, "Clarity")

class HelpfulnessRequest(SDKBaseModel):
    """
    Request type for helpfulness evaluation.
    :param query: The input query string.
    :param response: The response to evaluate.
    """
    query: str
    response: str

class HelpfulnessResponse(SDKBaseModel):
    """
    Response type for helpfulness evaluation.
    :param score: The overall helpfulness score.
    """
    score: float = Field(..., ge=0, le=100, description="Helpfulness score (0-100)")
    @field_validator("score", mode="before")
    @classmethod
    def check_score(cls, v: float) -> float:
        return cls.validate_score_range(v, 0, 100, "Helpfulness")

class FormalityRequest(SDKBaseModel):
    """
    Request type for formality evaluation.
    :param response: The response to evaluate.
    """
    response: str

class FormalityResponse(SDKBaseModel):
    """
    Response type for formality evaluation.
    :param score: The overall formality score.
    :param sentences: List of sentences analyzed.
    :param scores: List of scores for each sentence.
    """
    score: float = Field(..., ge=0, le=100, description="Formality score (0-100)")
    sentences: List[str]
    scores: List[float]
    @field_validator("score", mode="before")
    @classmethod
    def check_score(cls, v: float) -> float:
        return cls.validate_score_range(v, 0, 100, "Formality")
    @field_validator("scores", mode="before")
    @classmethod
    def check_scores(cls, v: List[float]) -> List[float]:
        return cls.validate_score_range(v, 0, 100, "Formality")

class SimplicityResponse(SDKBaseModel):
    """
    Response type for simplicity evaluation.
    :param score: The overall simplicity score (percentage).
    """
    score: float = Field(..., ge=0, le=100, description="Simplicity score (0-100)")
    @field_validator("score", mode="before")
    @classmethod
    def check_score(cls, v: float) -> float:
        return cls.validate_score_range(v, 0, 100, "Simplicity")

class SensitivityResponse(SDKBaseModel):
    """
    Response type for sensitivity evaluation.
    :param scores: Mapping of topic to sensitivity score.
    """
    scores: Dict[str, float]
    @field_validator("scores", mode="before")
    @classmethod
    def check_scores(cls, v: Dict[str, float]) -> Dict[str, float]:
        return cls.validate_score_range(v, 0, 100, "Sensitivity")

class ToxicityResponse(SDKBaseModel):
    """
    Response type for toxicity evaluation.
    :param labels: List of toxicity category labels.
    :param scores: List of scores for each label (0-100).
    """
    labels: List[str]
    scores: List[float]
    @field_validator("scores", mode="before")
    @classmethod
    def check_scores(cls, v: List[float]) -> List[float]:
        return cls.validate_score_range(v, 0, 100, "Toxicity")

class ToneResponse(SDKBaseModel):
    """
    Response type for tone evaluation.
    :param labels: List of detected tone labels.
    :param scores: List of scores for each label.
    """
    labels: List[str]
    scores: List[float]
    @field_validator("scores", mode="before")
    @classmethod
    def check_scores(cls, v: List[float]) -> List[float]:
        return cls.validate_score_range(v, 0, 100, "Tone")

class CostResponse(SDKBaseModel):
    """
    Response type for cost evaluation.
    :param cost_estimate_per_run: Estimated cost per run.
    :param total_project_cost_estimate: Total estimated cost for the project.
    """
    cost_estimate_per_run: float
    total_project_cost_estimate: float

class CarbonResponse(SDKBaseModel):
    """
    Response type for carbon emissions evaluation.
    :param carbon_emitted: Estimated carbon emitted (kg CO2e).
    :param sci_per_api_call: SCI per API call.
    :param sci_per_10k_calls: SCI per 10,000 calls.
    """
    carbon_emitted: float
    sci_per_api_call: float
    sci_per_10k_calls: float

class FaithfulnessRequest(SDKBaseModel):
    """
    Request type for faithfulness evaluation.
    :param query: The input query string.
    :param response: The response to evaluate.
    :param context: The context information (required).
    """
    query: str
    response: str
    context: Context

class AnswerRelevancyRequest(SDKBaseModel):
    """
    Request type for answer relevancy evaluation.
    :param query: The input query string.
    :param response: The response to evaluate.
    :param context: The context information (required).
    """
    query: str
    response: str
    context: Context

class SimplicityRequest(SDKBaseModel):
    """
    Request type for simplicity evaluation.
    :param response: The response to evaluate.
    """
    response: str

class SensitivityRequest(SDKBaseModel):
    """
    Request type for sensitivity evaluation.
    :param response: The response to evaluate.
    :param topics: List of topics to check for sensitivity.
    :param query: The input query string (optional).
    """
    response: str
    topics: List[str]
    query: Optional[str] = None

class ToxicityRequest(SDKBaseModel):
    """
    Request type for toxicity evaluation.
    :param query: The input query string.
    :param response: The response to evaluate.
    """
    query: str
    response: str

class ToneRequest(SDKBaseModel):
    """
    Request type for tone evaluation.
    :param response: The response to evaluate.
    """
    response: str

class CostRequest(SDKBaseModel):
    """
    Request type for cost evaluation.
    :param model_name: Name of the model (non-empty string).
    :param model_type: Type of the model (must be 'LLM' or 'Reranker').
    :param model_provider: Provider of the model (non-empty string).
    :param number_of_queries: Number of queries to estimate cost for (> 0).
    :param total_prompt_tokens: Total prompt tokens (> 0).
    :param total_completion_tokens: Total completion tokens (> 0).
    :param total_tokens: Total tokens (optional, > 0 if provided).
    :param instance_type: Instance type (optional).
    :param average_latency: Average latency (optional, > 0 if provided).
    """
    model_name: constr(strip_whitespace=True, min_length=1) = Field(..., description="Model name must be a non-empty string.")
    model_type: Literal["LLM", "Reranker"] = Field(..., description="Model type must be 'LLM' or 'Reranker'.")
    model_provider: constr(strip_whitespace=True, min_length=1) = Field(..., description="Model provider must be a non-empty string.")
    number_of_queries: PositiveInt = Field(..., description="Number of queries must be > 0.")
    total_prompt_tokens: PositiveInt = Field(..., description="Total number of prompt tokens (must be > 0).")
    total_completion_tokens: PositiveInt = Field(..., description="Total number of completion tokens (must be > 0).")
    total_tokens: Optional[PositiveInt] = Field(None, description="Total tokens (optional, must be > 0 if provided).")
    instance_type: Optional[str] = Field(None, description="Instance type required for Hugging Face models.")
    average_latency: Optional[PositiveFloat] = Field(None, description="Optional: Average latency in milliseconds, must be > 0 if provided.")

class CarbonRequest(SDKBaseModel):
    """
    Request type for carbon evaluation.
    :param processor_name: Name of the processor.
    :param provider_name: Name of the provider.
    :param provider_region: Region of the provider.
    :param instance_type: Instance type.
    :param average_latency: Average latency (ms).
    """
    processor_name: str
    provider_name: str
    provider_region: str
    instance_type: str
    average_latency: int

class GuardrailResponse(SDKBaseModel):
    """
    Response type for guardrail evaluation.
    :param passed: Whether all metrics passed.
    :param blocked: Whether the response is blocked due to failure.
    :param results: Dictionary of metric results, each containing 'passed' and 'result'.
    """
    passed: bool
    blocked: bool
    results: dict

    def to_json(self, **kwargs) -> str:
        """
        Return a JSON string representation of the guardrail evaluation, recursively serializing all nested SDK types.
        Use this for logging, API responses, or storage.
        """
        def serialize(obj: object) -> object:
            if hasattr(obj, "to_json"):
                return obj.model_dump()
            elif isinstance(obj, dict):
                return {k: serialize(v) for k, v in obj.items()}
            elif isinstance(obj, list):
                return [serialize(i) for i in obj]
            return obj
        data = {
            "passed": self.passed,
            "blocked": self.blocked,
            "results": serialize(self.results)
        }
        import json
        return json.dumps(data, **kwargs)

    def to_dict(self) -> dict:
        """
        Return a Python dict representation of the guardrail evaluation, recursively serializing all nested SDK types.
        Use this for programmatic access, further processing, or conversion to JSON via json.dumps().
        """
        def serialize(obj: object) -> object:
            if hasattr(obj, "to_json"):
                return obj.model_dump()
            elif isinstance(obj, dict):
                return {k: serialize(v) for k, v in obj.items()}
            elif isinstance(obj, list):
                return [serialize(i) for i in obj]
            return obj
        return {
            "passed": self.passed,
            "blocked": self.blocked,
            "results": serialize(self.results)
        } 



from typing import List, Optional

from trustwise.sdk.metrics.base import BaseMetric
from trustwise.sdk.types import HelpfulnessRequest, HelpfulnessResponse


class HelpfulnessMetric:
    """Helpfulness metric for evaluating response helpfulness."""
    def __init__(self, client) -> None:
        self.client = client
        self.base_url = client.config.get_alignment_url("v1")

    def evaluate(
        self,
        *,
        query: Optional[str] = None,
        response: Optional[str] = None,
        **kwargs
    ) -> HelpfulnessResponse:
        """
        Evaluate the helpfulness of a response to the query.

        Args:
            query: The query string (required)
            response: The response string (required)

        Returns:
            HelpfulnessResponse containing the evaluation results
        """
        req = BaseMetric.validate_request_model(HelpfulnessRequest, query=query, response=response, **kwargs)
        result = self.client._post(
            endpoint=f"{self.base_url}/helpfulness",
            data=req.to_dict()
        )
        return HelpfulnessResponse(**result)

    def batch_evaluate(
        self,
        inputs: List[HelpfulnessRequest]
    ) -> List[HelpfulnessResponse]:
        """Evaluate multiple inputs for helpfulness."""
        raise NotImplementedError("Batch evaluation not yet supported")

    def explain(
        self,
        *,
        query: str,
        response: str,
        **kwargs
    ) -> dict:
        """Get detailed explanation of the helpfulness evaluation."""
        # req = HelpfulnessRequest(query=query, response=response)
        raise NotImplementedError("Explanation not yet supported") 


from typing import Any, Dict, List, Optional, TypeVar

from trustwise.sdk.metrics.base import BaseSafetyMetric
from trustwise.sdk.metrics.safety.v3.metrics.answer_relevancy import (
    AnswerRelevancyMetric,
)
from trustwise.sdk.metrics.safety.v3.metrics.context_relevancy import (
    ContextRelevancyMetric,
)

# Import metric classes from metrics/ submodules
from trustwise.sdk.metrics.safety.v3.metrics.faithfulness import FaithfulnessMetric
from trustwise.sdk.metrics.safety.v3.metrics.pii import PIIMetric
from trustwise.sdk.metrics.safety.v3.metrics.prompt_injection import (
    PromptInjectionMetric,
)
from trustwise.sdk.metrics.safety.v3.metrics.summarization import SummarizationMetric
from trustwise.sdk.types import (
    AnswerRelevancyRequest,
    Context,
    ContextRelevancyRequest,
    FaithfulnessRequest,
    PIIRequest,
    PromptInjectionRequest,
    SummarizationRequest,
)

# Define generic type variables for request and response
TRequest = TypeVar("TRequest")
TResponse = TypeVar("TResponse")

class SafetyMetricsV3(BaseSafetyMetric):
    """Safety metrics implementation for version 3."""
    
    def __init__(self, client) -> None:
        super().__init__(
            client=client,
            base_url=client.config.get_safety_url("v3"),
            version="v3"
        )
        # Initialize specific metric implementations
        self.faithfulness = FaithfulnessMetric(client)
        self.answer_relevancy = AnswerRelevancyMetric(client)
        self.context_relevancy = ContextRelevancyMetric(client)
        self.summarization = SummarizationMetric(client)
        self.pii = PIIMetric(client)
        self.prompt_injection = PromptInjectionMetric(client)
        
        # Set up the registry with handler methods
        self._initialize_registry()

    def _initialize_registry(self) -> None:
        """Initialize the registry with all available metrics."""
        self._registry.register(FaithfulnessRequest, self.faithfulness.evaluate)
        self._registry.register(AnswerRelevancyRequest, self.answer_relevancy.evaluate)
        self._registry.register(ContextRelevancyRequest, self.context_relevancy.evaluate)
        self._registry.register(SummarizationRequest, self.summarization.evaluate)
        self._registry.register(PIIRequest, self.pii.evaluate)
        self._registry.register(PromptInjectionRequest, self.prompt_injection.evaluate)

    def batch_evaluate(self, inputs: List[Any]) -> List[Any]:
        """Evaluate multiple inputs in a single request.
        
        This method should be called with a list of specific metric request types.
        All requests in the list must be of the same type.
        """
        if not inputs:
            return []
            
        request_type = type(inputs[0])
        if not all(isinstance(x, request_type) for x in inputs):
            raise ValueError("All inputs must be of the same type")
            
        # Get the appropriate handler from the registry
        handler = self._registry.get_handler(inputs[0])
        if handler and callable(handler):
            # For now, still raise NotImplementedError since batch evaluation 
            # is not yet supported for any metrics
            raise NotImplementedError("Batch evaluation not yet supported")
        else:
            raise ValueError(f"Unsupported request type: {request_type}")

    def explain(self, query: str, response: str, context: Optional[Context] = None) -> Dict[str, Any]:
        """Get detailed explanation of the evaluation.
        
        This method provides explanations for all available metrics.
        """
        # Still returns empty dictionary since all metric.explain() methods will raise NotImplementedError
        return {} 


from typing import List, Optional

from trustwise.sdk.metrics.base import BaseMetric
from trustwise.sdk.types import CostRequest, CostResponse


class CostMetric:
    """Cost metrics evaluator."""
    def __init__(self, client) -> None:
        self.client = client
        self.base_url = client.config.get_performance_url("v1")

    def evaluate(
        self,
        *,
        model_name: Optional[str] = None,
        model_type: Optional[str] = None,
        model_provider: Optional[str] = None,
        number_of_queries: Optional[int] = None,
        total_prompt_tokens: Optional[int] = None,
        total_completion_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        instance_type: Optional[str] = None,
        average_latency: Optional[float] = None,
        **kwargs
    ) -> CostResponse:
        """
        Evaluate cost metrics.
        All arguments are required except those marked optional.
        """
        req = BaseMetric.validate_request_model(
            CostRequest,
            model_name=model_name,
            model_type=model_type,
            model_provider=model_provider,
            number_of_queries=number_of_queries,
            total_prompt_tokens=total_prompt_tokens,
            total_completion_tokens=total_completion_tokens,
            total_tokens=total_tokens,
            instance_type=instance_type,
            average_latency=average_latency,
            **kwargs
        )
        result = self.client._post(
            endpoint=f"{self.base_url}/cost",
            data=req.to_dict()
        )
        return CostResponse(**result)

    def batch_evaluate(
        self,
        inputs: List[CostRequest]
    ) -> List[CostResponse]:
        """Evaluate multiple inputs for cost."""
        raise NotImplementedError("Batch evaluation not yet supported")

    def explain(
        self,
        *,
        model_name: Optional[str] = None,
        model_type: Optional[str] = None,
        model_provider: Optional[str] = None,
        number_of_queries: Optional[int] = None,
        total_prompt_tokens: Optional[int] = None,
        total_completion_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        instance_type: Optional[str] = None,
        average_latency: Optional[float] = None,
        cost_map_name: str = "sys",
        **kwargs
    ) -> dict:
        """Get detailed explanation of the cost evaluation."""
        # req = CostRequest(
        #     model_name=model_name,
        #     model_type=model_type,
        #     model_provider=model_provider,
        #     number_of_queries=number_of_queries,
        #     total_prompt_tokens=total_prompt_tokens,
        #     total_completion_tokens=total_completion_tokens,
        #     total_tokens=total_tokens,
        #     instance_type=instance_type,
        #     average_latency=average_latency
        # )
        raise NotImplementedError("Explanation not yet supported") 


import logging
from abc import ABC, abstractmethod
from typing import (
    Any,
    Dict,
    Generic,
    List,
    Optional,
    Protocol,
    Type,
    TypeVar,
    runtime_checkable,
)
from urllib.parse import urljoin

from pydantic import ValidationError

from trustwise.sdk.exceptions import TrustwiseValidationError
from trustwise.sdk.types import (
    AnswerRelevancyRequest,
    AnswerRelevancyResponse,
    ClarityRequest,
    ClarityResponse,
    ContextRelevancyRequest,
    ContextRelevancyResponse,
    FaithfulnessRequest,
    FaithfulnessResponse,
    FormalityRequest,
    FormalityResponse,
    HelpfulnessRequest,
    HelpfulnessResponse,
    PIIRequest,
    PIIResponse,
    PromptInjectionRequest,
    PromptInjectionResponse,
    SDKBaseModel,
    SensitivityRequest,
    SensitivityResponse,
    SimplicityRequest,
    SimplicityResponse,
    SummarizationRequest,
    SummarizationResponse,
    ToxicityRequest,
    ToxicityResponse,
)

logger = logging.getLogger(__name__)

# Generic type variables for request and response
TRequest = TypeVar("TRequest")
TResponse = TypeVar("TResponse")


@runtime_checkable
class MetricProtocol(Protocol[TRequest, TResponse]):
    """Protocol defining the interface for all metrics."""
    def evaluate(self, request: TRequest) -> TResponse:
        """Evaluate the metric with a specific request."""
        ...
    
    def batch_evaluate(self, inputs: List[TRequest]) -> List[TResponse]:
        """Evaluate multiple inputs in a single request."""
        ...
    
    def explain(self, *args, **kwargs) -> Dict[str, Any]:
        """Get detailed explanation of the evaluation."""
        ...


class BaseMetric(ABC, Generic[TRequest, TResponse]):
    """Base class for all metrics."""
    
    def __init__(self, client, base_url: str, version: str) -> None:
        self.client = client
        self.base_url = base_url
        self.version = version
        logger.debug("Initialized %s with base_url: %s, version: %s", 
                    self.__class__.__name__, base_url, version)

    def _get_endpoint_url(self, endpoint: str) -> str:
        """Get the full URL for an endpoint."""
        url = urljoin(str(self.base_url) + "/", endpoint)
        logger.debug("Constructed URL for endpoint %s: %s", endpoint, url)
        return url

    @abstractmethod
    def evaluate(self, request: TRequest) -> TResponse:
        """Evaluate the metric. Must be implemented by concrete classes."""
        pass

    def batch_evaluate(self, inputs: List[TRequest]) -> List[TResponse]:
        """Evaluate multiple inputs in a single request. Optional implementation."""
        raise NotImplementedError("Batch evaluation not supported for this metric")

    def explain(self, *args, **kwargs) -> Dict[str, Any]:
        """Get detailed explanation of the evaluation. Optional implementation."""
        raise NotImplementedError("Explanation not supported for this metric")

    def _post(self, endpoint: str, data: object) -> object:
        """Make a POST request to the metric endpoint."""
        return self.client._post(self._get_endpoint_url(endpoint), data)

    @staticmethod
    def validate_request_model(model_cls, **kwargs) -> object:
        """
        Standardized Trustwise validation for all metric request models.
        Usage: req = BaseMetric.validate_request_model(RequestModel, **kwargs)
        Raises TrustwiseValidationError with a formatted message on error.
        """
        try:
            return model_cls(**kwargs)
        except ValidationError as ve:
            raise TrustwiseValidationError(SDKBaseModel.format_validation_error(model_cls, ve)) from ve
        except TypeError as te:
            # Detect missing required arguments
            import inspect
            sig = inspect.signature(model_cls)
            missing_args = []
            for name, param in sig.parameters.items():
                if param.default is param.empty and name not in kwargs:
                    missing_args.append(name)
            if missing_args:
                class DummyValidationError(Exception):
                    def errors(self) -> list:
                        return [
                            {"loc": [arg], "msg": "field required"} for arg in missing_args
                        ]
                ve = DummyValidationError()
                raise TrustwiseValidationError(SDKBaseModel.format_validation_error(model_cls, ve)) from te
            else:
                raise


class MetricEvaluator(BaseMetric, Generic[TRequest, TResponse]):
    """Base class for all metric evaluations with typed request/response."""
    
    def evaluate(self, request: TRequest) -> TResponse:
        """Evaluate the metric with a specific request type."""
        return self.client._post(
            endpoint=f"{self.base_url}/{self._get_endpoint_name()}",
            data=request
        )
    
    def batch_evaluate(self, inputs: List[TRequest]) -> List[TResponse]:
        """Evaluate multiple inputs in a single request."""
        raise NotImplementedError("Batch evaluation not yet supported")
    
    def explain(self, *args, **kwargs) -> Dict[str, Any]:
        """Get detailed explanation of the evaluation."""
        raise NotImplementedError("Explanation not yet supported")
    
    def _get_endpoint_name(self) -> str:
        """Get the endpoint name for this metric. Override in subclasses."""
        # Default implementation strips 'Metric' from the class name and converts to snake_case
        class_name = self.__class__.__name__
        metric_name = class_name[:-6] if class_name.endswith("Metric") else class_name
        return "".join(["_" + c.lower() if c.isupper() else c for c in metric_name]).lstrip("_")


class MetricRegistry:
    """Registry for mapping request types to metric handlers."""
    
    def __init__(self) -> None:
        self._registry: Dict[Type, object] = {}
    
    def register(self, request_type: Type, handler: object) -> None:
        """Register a handler for a specific request type."""
        self._registry[request_type] = handler
    
    def get_handler(self, request) -> Optional[object]:
        """Get the appropriate handler for a given request."""
        request_type = type(request)
        return self._registry.get(request_type)


class BaseSafetyMetric(BaseMetric[Any, Any]):
    """Base class for safety metrics."""
    
    def __init__(self, client, base_url: str, version: str) -> None:
        super().__init__(client, base_url, version)
        self._registry = MetricRegistry()
        self._initialize_registry()
    
    def _initialize_registry(self) -> None:
        """Initialize the registry with available metrics. Override in concrete classes."""
        pass
        
    def evaluate(self, request) -> object:
        """Evaluate the safety metrics based on request type."""
        handler = self._registry.get_handler(request)
        if handler:
            return handler(request)
        else:
            raise ValueError(f"Unsupported request type: {type(request)}")

    def batch_evaluate(self, inputs: List[Any]) -> List[Any]:
        """Evaluate multiple inputs in a single request."""
        raise NotImplementedError("Batch evaluation not yet supported")

    def explain(self, *args, **kwargs) -> Dict[str, Any]:
        """Get detailed explanation of the evaluation."""
        raise NotImplementedError("Explanation not yet supported")
    
    def faithfulness(self, request: FaithfulnessRequest) -> FaithfulnessResponse:
        """Evaluate faithfulness of a response."""
        return self.client._post(self._get_endpoint_url("faithfulness"), request)
    
    def context_relevancy(self, request: ContextRelevancyRequest) -> ContextRelevancyResponse:
        """Evaluate context relevancy of a response."""
        return self.client._post(self._get_endpoint_url("context_relevancy"), request)
    
    def answer_relevancy(self, request: AnswerRelevancyRequest) -> AnswerRelevancyResponse:
        """Evaluate answer relevancy of a response."""
        return self.client._post(self._get_endpoint_url("answer_relevancy"), request)
    
    def summarization(self, request: SummarizationRequest) -> SummarizationResponse:
        """Evaluate summarization quality."""
        return self.client._post(self._get_endpoint_url("summarization"), request)
    
    def clarity(self, request: ClarityRequest) -> ClarityResponse:
        """Evaluate clarity of a response."""
        return self.client._post(self._get_endpoint_url("clarity"), request)
    
    def helpfulness(self, request: HelpfulnessRequest) -> HelpfulnessResponse:
        """Evaluate helpfulness of a response."""
        return self.client._post(self._get_endpoint_url("helpfulness"), request)

    def pii(self, request: PIIRequest) -> PIIResponse:
        """Evaluate text for PII content."""
        if not request.text or not isinstance(request.text, str):
            raise ValueError("Text must be a non-empty string")
            
        return self.client._post(self._get_endpoint_url("pii"), request)

    def prompt_injection(self, request: PromptInjectionRequest) -> PromptInjectionResponse:
        """Evaluate text for prompt injection attempts."""
        if not request.query or not isinstance(request.query, str):
            raise ValueError("Query must be a non-empty string")
            
        return self.client._post(self._get_endpoint_url("prompt_injection"), request)


class BaseAlignmentMetric(BaseMetric[Any, Any]):
    """Base class for alignment metrics."""
    
    def __init__(self, client, base_url: str, version: str) -> None:
        super().__init__(client, base_url, version)
        self._registry = MetricRegistry()
        # Do NOT call self._initialize_registry() here; let subclasses call it after their own init
    
    def evaluate(self, request) -> object:
        """Evaluate the alignment metrics based on request type."""
        handler = self._registry.get_handler(request)
        if handler:
            return handler(request)
        else:
            # Default to clarity if not found
            if isinstance(request, ClarityRequest):
                return self.clarity(request)
            else:
                raise ValueError(f"Unsupported request type: {type(request)}")

    def batch_evaluate(self, inputs: List[Any]) -> List[Any]:
        """Evaluate multiple inputs in a single request."""
        raise NotImplementedError("Batch evaluation not yet supported")

    def explain(self, *args, **kwargs) -> Dict[str, Any]:
        """Get detailed explanation of the evaluation."""
        raise NotImplementedError("Explanation not yet supported")
    
    def clarity(self, request: ClarityRequest) -> ClarityResponse:
        """Evaluate clarity of a response."""
        return self.client._post(self._get_endpoint_url("clarity"), request)
    
    def helpfulness(self, request: HelpfulnessRequest) -> HelpfulnessResponse:
        """Evaluate helpfulness of a response."""
        return self.client._post(self._get_endpoint_url("helpfulness"), request)
    
    def formality(self, request: FormalityRequest) -> FormalityResponse:
        """Evaluate formality of a response."""
        return self.client._post(self._get_endpoint_url("formality"), request)
    
    def simplicity(self, request: SimplicityRequest) -> SimplicityResponse:
        """Evaluate simplicity of a response."""
        return self.client._post(self._get_endpoint_url("simplicity"), request)
    
    def sensitivity(self, request: SensitivityRequest) -> SensitivityResponse:
        """Evaluate sensitivity regarding specific topics."""
        return self.client._post(self._get_endpoint_url("sensitivity"), request)
    
    def toxicity(self, request: ToxicityRequest) -> ToxicityResponse:
        """Evaluate toxicity."""
        return self.client._post(self._get_endpoint_url("toxicity"), request)


class BasePerformanceMetric(BaseMetric[Dict[str, Any], Dict[str, Any]]):
    """Base class for performance metrics."""
    
    def __init__(self, client, base_url: str, version: str) -> None:
        super().__init__(client, base_url, version)
        self._registry = MetricRegistry()
        # Do NOT call self._initialize_registry() here; let subclasses call it after their own init
    
    def evaluate(self, request: Dict[str, Any]) -> Dict[str, Any]:
        """Evaluate the performance metrics."""
        # Default to using cost metric with the request
        return self.cost(**request)

    def batch_evaluate(self, inputs: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Evaluate multiple inputs in a single request."""
        raise NotImplementedError("Batch evaluation not yet supported")

    def explain(self, *args, **kwargs) -> Dict[str, Any]:
        """Get detailed explanation of the evaluation."""
        raise NotImplementedError("Explanation not yet supported")
    
    def cost(
        self,
        total_prompt_tokens: int,
        total_completion_tokens: int,
        model_name: str,
        model_provider: str,
        average_latency: int,
        number_of_queries: int,
        instance_type: str
    ) -> Dict[str, Any]:
        """Evaluate cost metrics."""
        data = {
            "total_prompt_tokens": total_prompt_tokens,
            "total_completion_tokens": total_completion_tokens,
            "model_name": model_name,
            "model_provider": model_provider,
            "average_latency": average_latency,
            "number_of_queries": number_of_queries,
            "instance_type": instance_type
        }
        return self.client._post(self._get_endpoint_url("cost"), data)
    
    def carbon(
        self,
        processor_name: str,
        provider_name: str,
        provider_region: str,
        instance_type: str,
        average_latency: int
    ) -> Dict[str, Any]:
        """Evaluate carbon metrics."""
        data = {
            "processor_name": processor_name,
            "provider_name": provider_name,
            "provider_region": provider_region,
            "instance_type": instance_type,
            "average_latency": average_latency
        }
        return self.client._post(self._get_endpoint_url("carbon"), data) 


"""
Simple example of using Trustwise SDK with LangChain.

This example demonstrates a basic integration with LangChain
using the document about tire maintenance.
"""

import os
import traceback
from typing import Any, Dict, List

from langchain_core.documents import Document
from langchain_core.output_parsers import StrOutputParser

# Import LangChain components
from langchain_core.prompts import PromptTemplate
from langchain_core.runnables import RunnablePassthrough
from langchain_openai import ChatOpenAI
from langchain_text_splitters import RecursiveCharacterTextSplitter

# Import Trustwise SDK - Fix the import path based on your structure
from trustwise import Trustwise

# Set your OpenAI API key
# os.environ["OPENAI_API_KEY"] = "your-openai-api-key"

# Set your Trustwise API key
# os.environ["TRUSTWISE_API_KEY"] = "your-trustwise-api-key"


def run_trustwise_langchain_example():
    """Run the Trustwise LangChain example"""

    print("Trustwise SDK with LangChain Example")
    print("------------------------------------")

    # Load the tire maintenance document
    current_dir = os.path.dirname(os.path.abspath(__file__))
    project_root = os.path.dirname(os.path.dirname(os.path.dirname(current_dir)))
    document_path = os.path.join(current_dir, "tire_maintenance.txt")

    # Create sample document if it doesn't exist
    if not os.path.exists(document_path):
        sample_text = """
        # Tire Maintenance Guide
        
        ## Checking Tire Pressure
        
        To check tire pressure, you need a tire pressure gauge. These can be purchased at auto parts stores for $1-7.
        
        Follow these steps:
        1. Remove the valve cap
        2. Press the gauge onto the valve firmly
        3. Read the pressure measurement
        4. Add air if needed and recheck
        
        Proper tire pressure improves fuel efficiency and extends tire life.
        
        The recommended tire pressure for your vehicle can be found:
        - On a sticker in the driver's door jamb
        - In the vehicle owner's manual
        - Sometimes on the inside of the fuel filler door
        
        Most passenger vehicles have a recommended pressure between 32-35 PSI.
        """
        with open(document_path, "w") as f:
            f.write(sample_text)
        print(f"Created sample document at {document_path}")

    with open(document_path, "r") as f:
        document_content = f.read()

    # Create Document object
    document = Document(page_content=document_content)

    # Initialize Trustwise
    trustwise_api_key = os.environ.get("TRUSTWISE_API_KEY")
    if not trustwise_api_key:
        print(
            "Warning: TRUSTWISE_API_KEY environment variable not set. Using placeholder."
        )
        trustwise_api_key = (
            "your-trustwise-api-key"  # This won't work with actual API calls
        )

    trustwise = Trustwise(api_key=trustwise_api_key)

    # Get the Trustwise callback handler for LangChain - using the correct path
    try:
        # Try three different import paths based on your structure
        try:
            from trustwise.integrations import get_langchain_callback_handler
        except ImportError:
            try:
                from trustwise import integrations

                get_langchain_callback_handler = (
                    integrations.get_langchain_callback_handler
                )
            except (ImportError, AttributeError):
                from trustwise import get_langchain_callback_handler

        TrustwiseCallbackHandler = get_langchain_callback_handler()

        if TrustwiseCallbackHandler is None:
            raise ImportError(
                "TrustwiseCallbackHandler is None - LangChain may not be installed correctly"
            )

        print("Successfully imported TrustwiseCallbackHandler")
    except Exception as e:
        print(f"Error importing callback handler: {str(e)}")
        traceback.print_exc()
        return

    # Create thresholds for metrics
    thresholds = {"clarity": 70.0, "helpfulness": 70.0}

    # Create callback function for threshold failures
    def on_threshold_failure(metric, result):
        print(
            f"⚠️ Warning: {metric} score {result['score']:.2f} is below threshold {thresholds[metric]}"
        )

    # Create debug wrapper for the callback handler
    class DebugTrustwiseCallbackHandler(TrustwiseCallbackHandler):
        def on_chain_start(
            self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs
        ):
            print(f"[Debug] Chain Start triggered with inputs: {list(inputs.keys())}")
            super().on_chain_start(serialized, inputs, **kwargs)
            print(f"[Debug] Current query: {getattr(self, '_current_query', None)}")
            print(
                f"[Debug] Current context: {getattr(self, '_current_context', None) is not None}"
            )

        def on_chain_end(self, outputs: Dict[str, Any], **kwargs):
            print(f"[Debug] Chain End triggered with outputs: {list(outputs.keys())}")
            if "output" in outputs:
                print(f"[Debug] Output type: {type(outputs['output'])}")
                print(f"[Debug] Output preview: {str(outputs['output'])[:50]}...")

            super().on_chain_end(outputs, **kwargs)
            print(f"[Debug] Latest results: {getattr(self, 'latest_results', None)}")

    # Initialize callback handler
    trustwise_handler = DebugTrustwiseCallbackHandler(
        trustwise_client=trustwise,
        metrics=["clarity", "helpfulness"],
        threshold_config=thresholds,
        on_threshold_failure=on_threshold_failure,
    )

    # Create a simple retrieval function that returns our document
    def retrieve_docs(query: str) -> List[Document]:
        return [document]

    # Create LLM
    openai_api_key = os.environ.get("OPENAI_API_KEY")
    if not openai_api_key:
        print("Warning: OPENAI_API_KEY environment variable not set.")
        print("Example will run in simulated mode without making real API calls.")

        # Create a mock chain for testing without API keys
        def mock_chain(input_text, config=None):
            query = input_text
            if query.lower().startswith("how do i check"):
                return "To check tire pressure, you need a pressure gauge. Remove the valve cap, press the gauge onto the valve firmly, and read the pressure."
            elif query.lower().startswith("what tools"):
                return "The main tool you need is a tire pressure gauge, which can be purchased at auto parts stores for $1-7."
            elif query.lower().startswith("where can i find"):
                return "You can find the recommended tire pressure on a sticker in the driver's door jamb, in the vehicle owner's manual, or sometimes on the inside of the fuel filler door."
            else:
                return "I don't have specific information about that question."

        chain = mock_chain
    else:
        llm = ChatOpenAI(temperature=0.2)

        # Create prompt template
        prompt = PromptTemplate.from_template(
            "Answer the following question based on the context:\n\n"
            "Context: {context}\n\n"
            "Question: {input}\n\n"
            "Answer:"
        )

        # Create simple chain - fixed to work with current LangChain
        chain = (
            {
                "context": lambda x: "\n".join(
                    [doc.page_content for doc in retrieve_docs(x)]
                ),
                "input": RunnablePassthrough(),
            }
            | prompt
            | llm
            | StrOutputParser()
        )

    # Ask some questions
    questions = [
        "How do I check tire pressure?",
        "What tools do I need to check tire pressure?",
        "Where can I find the recommended tire pressure for my car?",
    ]

    for i, question in enumerate(questions, 1):
        print(f"\n[Question {i}] {question}")
        try:
            # Run chain with Trustwise evaluation
            if callable(getattr(chain, "invoke", None)):
                # For LangChain's Runnable protocol
                response = chain.invoke(
                    question, config={"callbacks": [trustwise_handler]}
                )
            else:
                # For the mock chain
                response = chain(question, config={"callbacks": [trustwise_handler]})

                # Manually trigger callbacks since we're using a mock
                if hasattr(trustwise_handler, "on_chain_start"):
                    trustwise_handler.on_chain_start(
                        serialized={}, inputs={"input": question}
                    )

                if hasattr(trustwise_handler, "on_chain_end"):
                    trustwise_handler.on_chain_end(outputs={"output": response})

            print(f"[Response] {response}")

            # Print evaluation results with more detailed logging
            print("[Evaluation Results]")
            if hasattr(trustwise_handler, "latest_results"):
                if not trustwise_handler.latest_results:
                    print("No evaluation results available")
                else:
                    for metric, result in trustwise_handler.latest_results.items():
                        if isinstance(result, dict) and "score" in result:
                            print(f"- {metric.capitalize()}: {result['score']:.2f}")
                        else:
                            print(
                                f"- {metric.capitalize()}: Result format unexpected - {result}"
                            )
            else:
                print("No latest_results attribute found in handler")

        except Exception as e:
            print(f"Error during evaluation: {str(e)}")
            traceback.print_exc()

        print("-" * 40)


if __name__ == "__main__":
    run_trustwise_langchain_example()



class TrustwiseSDKError(Exception):
    """Base exception for all Trustwise SDK errors."""
    pass

class TrustwiseValidationError(TrustwiseSDKError):
    """Exception for validation errors in Trustwise SDK."""
    pass 

class TrustwiseAPIError(TrustwiseSDKError):
    def __init__(self, message, response, status_code) -> None:
        super().__init__(message)
        self.response = response
        self.status_code = status_code

def make_status_error_from_response(response) -> "TrustwiseAPIError":
    """
    Create a TrustwiseAPIError from a requests.Response
    """
    try:
        err_text = response.text.strip()
        body = err_text
        try:
            body = response.json()
            err_msg = f"Error code: {response.status_code} - {body}"
        except ValueError:
            err_msg = err_text or f"Error code: {response.status_code}"
    except AttributeError:
        body = None
        err_msg = f"Error code: {response.status_code}"
    return TrustwiseAPIError(err_msg, body, response.status_code)



from typing import List, Optional

from trustwise.sdk.metrics.base import BaseMetric
from trustwise.sdk.types import ClarityRequest, ClarityResponse


class ClarityMetric:
    """Clarity metric for evaluating response clarity."""
    def __init__(self, client) -> None:
        self.client = client
        self.base_url = client.config.get_alignment_url("v1")

    def evaluate(
        self,
        *,
        query: Optional[str] = None,
        response: Optional[str] = None,
        **kwargs
    ) -> ClarityResponse:
        """
        Evaluate the clarity of a response to the query.

        Args:
            query: The query string (required)
            response: The response string (required)

        Returns:
            ClarityResponse containing the evaluation results
        """
        req = BaseMetric.validate_request_model(ClarityRequest, query=query, response=response, **kwargs)
        result = self.client._post(
            endpoint=f"{self.base_url}/clarity",
            data=req.to_dict()
        )
        return ClarityResponse(**result)

    def batch_evaluate(
        self,
        inputs: List[ClarityRequest]
    ) -> List[ClarityResponse]:
        """Evaluate multiple inputs for clarity."""
        raise NotImplementedError("Batch evaluation not yet supported")

    def explain(
        self,
        *,
        query: str,
        response: str,
        **kwargs
    ) -> dict:
        """Get detailed explanation of the clarity evaluation."""
        # req = ClarityRequest(query=query, response=response)
        raise NotImplementedError("Explanation not yet supported") 


from trustwise.sdk.metrics.alignment.v1.metrics.clarity import ClarityMetric
from trustwise.sdk.metrics.alignment.v1.metrics.formality import FormalityMetric
from trustwise.sdk.metrics.alignment.v1.metrics.helpfulness import HelpfulnessMetric
from trustwise.sdk.metrics.alignment.v1.metrics.sensitivity import SensitivityMetric
from trustwise.sdk.metrics.alignment.v1.metrics.simplicity import SimplicityMetric
from trustwise.sdk.metrics.alignment.v1.metrics.tone import ToneMetric
from trustwise.sdk.metrics.alignment.v1.metrics.toxicity import ToxicityMetric
from trustwise.sdk.metrics.base import BaseAlignmentMetric


class AlignmentMetricsV1(BaseAlignmentMetric):
    """Alignment metrics for version 1 of the API."""
    
    def __init__(self, client) -> None:
        """Initialize the safety metrics."""
        base_url = client.config.get_safety_url("v3")
        super().__init__(client, base_url=base_url, version="v3")
        self.clarity = ClarityMetric(client)
        self.helpfulness = HelpfulnessMetric(client)
        self.formality = FormalityMetric(client)
        self.simplicity = SimplicityMetric(client)
        self.tone = ToneMetric(client)
        self.toxicity = ToxicityMetric(client)
        self.sensitivity = SensitivityMetric(client)



"""Unit tests for Trustwise configuration."""

import os
from unittest.mock import patch
from urllib.parse import urlparse

import pytest

from trustwise.sdk.config import TrustwiseConfig


class TestTrustwiseConfig:
    """Test cases for Trustwise configuration."""
    
    def test_default_base_url(self):
        """Test default base URL when no value is provided."""
        config = TrustwiseConfig(api_key="test-key")
        assert config.base_url == "https://api.trustwise.ai/"
    
    def test_custom_base_url(self):
        """Test custom base URL."""
        # Test with trailing slash
        custom_url = "https://custom.trustwise.ai/"
        config = TrustwiseConfig(api_key="test-key", base_url=custom_url)
        assert config.base_url == custom_url
        
        # Test without trailing slash
        custom_url = "https://custom.trustwise.ai"
        config = TrustwiseConfig(api_key="test-key", base_url=custom_url)
        assert config.base_url == f"{custom_url}/"
    
    def test_base_url_from_env(self):
        """Test base URL from environment variable."""
        # Test with trailing slash
        custom_url = "https://env.trustwise.ai/"
        with patch.dict(os.environ, {"TW_BASE_URL": custom_url}):
            config = TrustwiseConfig(api_key="test-key")
            assert config.base_url == custom_url
        
        # Test without trailing slash
        custom_url = "https://env.trustwise.ai"
        with patch.dict(os.environ, {"TW_BASE_URL": custom_url}):
            config = TrustwiseConfig(api_key="test-key")
            assert config.base_url == f"{custom_url}/"
    
    def test_api_key_required_validation(self):
        """Test API key validation."""
        # Test missing API key
        with patch.dict(os.environ, {}, clear=True):
            with pytest.raises(ValueError) as exc_info:
                TrustwiseConfig()
            assert "API key must be provided" in str(exc_info.value)
        
        # Test empty API key
        with pytest.raises(ValueError) as exc_info:
            TrustwiseConfig(api_key="")
        assert "API key must be provided" in str(exc_info.value)
        
        # Test None API key
        with pytest.raises(ValueError) as exc_info:
            TrustwiseConfig(api_key=None)
        assert "API key must be provided either directly or through TW_API_KEY environment variable" in str(exc_info.value)
        
        # Test invalid API key type
        with pytest.raises(ValueError) as exc_info:
            TrustwiseConfig(api_key=123)  # type: ignore
        assert "API key must be a valid string" in str(exc_info.value)
    
    def test_api_key_from_env(self):
        """Test API key from environment variable."""
        with patch.dict(os.environ, {"TW_API_KEY": "env-key"}):
            config = TrustwiseConfig()
            assert config.api_key == "env-key"
    
    def test_api_key_precedence(self):
        """Test that direct API key takes precedence over environment variable."""
        with patch.dict(os.environ, {"TW_API_KEY": "env-key"}):
            config = TrustwiseConfig(api_key="direct-key")
            assert config.api_key == "direct-key"
    
    def test_base_url_validation(self):
        """Test base URL validation."""
        # Test invalid base URL format
        with pytest.raises(ValueError) as exc_info:
            TrustwiseConfig(api_key="test-key", base_url="invalid-url")
        assert "Invalid base URL format" in str(exc_info.value)
        
        # Test invalid base URL type
        with pytest.raises(ValueError) as exc_info:
            TrustwiseConfig(api_key="test-key", base_url=123)  # type: ignore
        assert "Base URL must be a valid string" in str(exc_info.value)
        
        # Test missing scheme
        with pytest.raises(ValueError) as exc_info:
            TrustwiseConfig(api_key="test-key", base_url="trustwise.ai")
        assert "Invalid base URL format" in str(exc_info.value)
        
        # Test missing netloc
        with pytest.raises(ValueError) as exc_info:
            TrustwiseConfig(api_key="test-key", base_url="https://")
        assert "Invalid base URL format" in str(exc_info.value)
    
    def test_get_safety_url(self):
        """Test safety URL construction."""
        # Test with default base URL
        config = TrustwiseConfig(api_key="test-key")
        url = config.get_safety_url("v3")
        assert url == "https://api.trustwise.ai/safety/v3"
        assert urlparse(url).scheme == "https"
        
        # Test with custom base URL
        custom_url = "https://custom.trustwise.ai/api"
        config = TrustwiseConfig(api_key="test-key", base_url=custom_url)
        url = config.get_safety_url("v3")
        assert url == "https://custom.trustwise.ai/api/safety/v3"
    
    def test_get_alignment_url(self):
        """Test alignment URL construction."""
        # Test with default base URL
        config = TrustwiseConfig(api_key="test-key")
        url = config.get_alignment_url("v1")
        assert url == "https://api.trustwise.ai/alignment/v1"
        assert urlparse(url).scheme == "https"
        
        # Test with custom base URL
        custom_url = "https://custom.trustwise.ai/api"
        config = TrustwiseConfig(api_key="test-key", base_url=custom_url)
        url = config.get_alignment_url("v1")
        assert url == "https://custom.trustwise.ai/api/alignment/v1"
    
    def test_custom_base_url_with_paths(self):
        """Test URL construction with custom base URL."""
        # Test with trailing slash
        custom_url = "https://custom.trustwise.ai/api/"
        config = TrustwiseConfig(api_key="test-key", base_url=custom_url)
        assert config.get_safety_url("v3") == "https://custom.trustwise.ai/api/safety/v3"
        assert config.get_alignment_url("v1") == "https://custom.trustwise.ai/api/alignment/v1"
        
        # Test without trailing slash
        custom_url = "https://custom.trustwise.ai/api"
        config = TrustwiseConfig(api_key="test-key", base_url=custom_url)
        assert config.get_safety_url("v3") == "https://custom.trustwise.ai/api/safety/v3"
        assert config.get_alignment_url("v1") == "https://custom.trustwise.ai/api/alignment/v1"

def test_post_raises_trustwise_api_error_on_non_200(monkeypatch):
    import requests

    from trustwise.sdk.client import TrustwiseClient
    from trustwise.sdk.config import TrustwiseConfig
    from trustwise.sdk.exceptions import TrustwiseAPIError

    class MockResponse:
        def __init__(self, status_code, json_data=None, text_data=None):
            self.status_code = status_code
            self._json_data = json_data
            self.text = text_data or ""
            self.headers = {}
            self.raw = type("raw", (), {"closed": False})()
        def json(self):
            if self._json_data is not None:
                return self._json_data
            raise ValueError("No JSON")

    def mock_post(*args, **kwargs):
        return MockResponse(401, json_data={"message": "Auth error"}, text_data='{"message": "Auth error"}')

    monkeypatch.setattr(requests, "post", mock_post)
    config = TrustwiseConfig(api_key="bad", base_url="https://api.example.com/")
    client = TrustwiseClient(config)
    with pytest.raises(TrustwiseAPIError) as exc_info:
        client._post("https://api.example.com/test", data={})
    assert exc_info.value.status_code == 401
    assert exc_info.value.response == {"message": "Auth error"}
    assert "Auth error" in str(exc_info.value)


def test_post_raises_trustwise_api_error_on_non_json(monkeypatch):
    import requests

    from trustwise.sdk.client import TrustwiseClient
    from trustwise.sdk.config import TrustwiseConfig
    from trustwise.sdk.exceptions import TrustwiseAPIError

    class MockResponse:
        def __init__(self, status_code, text_data=None):
            self.status_code = status_code
            self.text = text_data or ""
            self.headers = {}
            self.raw = type("raw", (), {"closed": False})()
        def json(self):
            raise ValueError("No JSON")

    def mock_post(*args, **kwargs):
        return MockResponse(500, text_data="Internal Server Error")

    monkeypatch.setattr(requests, "post", mock_post)
    config = TrustwiseConfig(api_key="bad", base_url="https://api.example.com/")
    client = TrustwiseClient(config)
    with pytest.raises(TrustwiseAPIError) as exc_info:
        client._post("https://api.example.com/test", data={})
    assert exc_info.value.status_code == 500
    assert exc_info.value.response == "Internal Server Error"
    assert "Internal Server Error" in str(exc_info.value) 


"""Tests to verify package installation and functionality."""

import os
import subprocess
import sys
import tempfile
from pathlib import Path
from typing import List, Optional

import pytest


def safe_subprocess_run(
    cmd: List[str],
    *,
    capture_output: bool = True,
    text: bool = True,
    check: bool = True,
    env: Optional[dict] = None,
) -> subprocess.CompletedProcess:
    """Safely run a subprocess command with proper security settings.
    
    Args:
        cmd: List of command arguments
        capture_output: Whether to capture stdout and stderr
        text: Whether to return output as text
        check: Whether to check return code
        env: Environment variables to use
        
    Returns:
        CompletedProcess instance
        
    Raises:
        subprocess.CalledProcessError: If check=True and process returns non-zero
    """
    return subprocess.run(  # noqa: S603
        cmd,
        capture_output=capture_output,
        text=text,
        check=check,
        shell=False,  # Never use shell=True for security
        env=env,
    )


def test_package_installation():
    """Test that the package can be installed via pip."""
    # Create a temporary directory for the virtual environment
    with tempfile.TemporaryDirectory() as temp_dir:
        venv_path = Path(temp_dir) / "venv"
        dist_dir = Path(temp_dir) / "dist"
        
        # Create a virtual environment
        safe_subprocess_run([sys.executable, "-m", "venv", str(venv_path)])
        
        # Get the python executable path
        if os.name == "nt":  # Windows
            python_path = venv_path / "Scripts" / "python"
        else:  # Unix-like
            python_path = venv_path / "bin" / "python"
        
        # Install build tools in the virtual environment
        # We need to download these wheels first and store them in the repo
        build_wheel = Path("tests/wheels/build-1.0.3-py3-none-any.whl")
        if not build_wheel.exists():
            pytest.skip("Build wheel not found. Please download it and place in tests/wheels/")
        
        safe_subprocess_run(
            [str(python_path), "-m", "pip", "install", "--no-index", str(build_wheel)]
        )
        
        # Build the distribution package
        safe_subprocess_run(
            [str(python_path), "-m", "build", "--outdir", str(dist_dir)]
        )
        
        # Find the wheel file
        wheel_files = list(dist_dir.glob("*.whl"))
        assert len(wheel_files) == 1, "Expected exactly one wheel file"
        wheel_file = wheel_files[0]
        
        # Install the package from the wheel without dependencies
        safe_subprocess_run(
            [str(python_path), "-m", "pip", "install", "--no-deps", str(wheel_file)]
        )
        
        # Install dependencies from local wheels
        wheels_dir = Path("tests/wheels")
        if not wheels_dir.exists():
            pytest.skip("Dependency wheels not found. Please download them and place in tests/wheels/")
        
        # Install each dependency wheel
        for wheel in wheels_dir.glob("*.whl"):
            if wheel.name != build_wheel.name:  # Skip the build wheel
                safe_subprocess_run(
                    [str(python_path), "-m", "pip", "install", "--no-index", str(wheel)]
                )
        
        # Verify the package is installed
        result = safe_subprocess_run(
            [str(python_path), "-m", "pip", "list"]
        )
        assert "trustwise" in result.stdout
        
        # Test importing the package with a clean Python environment
        env = os.environ.copy()
        env["PYTHONPATH"] = ""  # Clear PYTHONPATH to ensure isolation
        env["PYTHONHOME"] = ""  # Clear PYTHONHOME to ensure isolation
        
        # Test importing the package
        safe_subprocess_run(
            [str(python_path), "-c", "import trustwise"],
            env=env
        )
        
        # Test importing specific modules
        safe_subprocess_run(
            [str(python_path), "-c", "from trustwise.sdk import TrustwiseSDK"],
            env=env
        )
        
        # Verify no system packages are being used
        try:
            result = safe_subprocess_run(
                [str(python_path), "-c", "import sys; print('\\n'.join(sys.path))"],
                env=env
            )
            # Ensure no system paths are in sys.path
            paths = result.stdout.strip().split("\n")
            assert not any("site-packages" in path for path in paths), "Found system site-packages in sys.path"
            assert any(str(venv_path) in path for path in paths), "Virtual environment path not found in sys.path"
        except subprocess.CalledProcessError as e:
            pytest.fail(f"Failed to check sys.path: {e.stderr}")


def test_package_metadata():
    """Test that package metadata is correctly set."""
    import trustwise
    
    # Test version
    assert hasattr(trustwise, "__version__")
    assert isinstance(trustwise.__version__, str)
    
    # Test package name
    assert trustwise.__name__ == "trustwise"
    
    # Test that main classes are importable
    from trustwise.sdk import TrustwiseSDK
    from trustwise.sdk.config import TrustwiseConfig
    
    assert TrustwiseSDK is not None
    assert TrustwiseConfig is not None 


"""Test helper functions and mock data."""
from typing import Any, Dict

# Mock API responses for safety metrics
MOCK_SAFETY_RESPONSES: Dict[str, Dict[str, Any]] = {
    "faithfulness": {
        "score": 95.5,
        "facts": [
            {
                "statement": "The capital of France is Paris",
                "label": "SUPPORTED",
                "prob": 0.95,
                "sentence_span": [0, 42]
            }
        ]
    },
    "answer_relevancy": {
        "score": 92.0,
        "generated_question": "What is the capital of France?"
    },
    "context_relevancy": {
        "score": 88.5,
        "topics": ["geography", "cities"],
        "scores": [0.9, 0.85]
    },
    "pii": {
        "identified_pii": [
            {
                "string": "john@example.com",
                "category": "EMAIL",
                "interval": [0, 15]
            }
        ]
    },
    "summarization": {
        "score": 90.0
    },
    "prompt_injection": {
        "score": 99.887375
    }
}

# Mock API responses for alignment metrics
MOCK_ALIGNMENT_RESPONSES: Dict[str, Dict[str, Any]] = {
    "clarity": {
        "score": 90.0
    },
    "helpfulness": {
        "score": 85.0
    },
    "formality": {
        "score": 75.0,
        "sentences": [
            "Nuclear fusion is the process by which two or more protons and neutrons combine to form a single nucleus."
        ],
        "scores": [75.0]
    },
    "simplicity": {
        "score": 78.61597
    },
    "sensitivity": {
        "scores": {
            "health": 95.0,
            "finance": 85.0
        }
    },
    "toxicity": {
        "labels": ["identity_hate", "obscene", "threat", "insult", "toxic"],
        "scores": [0.036089644, 0.105483316, 0.027964465, 0.06207772, 0.3622106]
    },
    "tone": {
        "labels": [
            "neutral",
            "happiness",
            "realization"
        ],
        "scores": [
            89.704185,
            6.6798472,
            2.9873204
        ]
    }
}

# Mock API responses for performance metrics
MOCK_PERFORMANCE_RESPONSES: Dict[str, Dict[str, Any]] = {
    "cost": {
        "cost_estimate_per_run": 0.00055,
        "total_project_cost_estimate": 1.1
    },
    "carbon": {
        "carbon_emitted": 0.11408193333333336,
        "sci_per_api_call": 0.0005714199095626406,
        "sci_per_10k_calls": 5.714199095626406
    }
}

def get_mock_response(endpoint: str) -> Dict[str, Any]:
    """Get mock response for a given API endpoint."""
    if "safety" in endpoint:
        metric = endpoint.split("/")[-1]
        return MOCK_SAFETY_RESPONSES.get(metric, {})
    elif "alignment" in endpoint:
        metric = endpoint.split("/")[-1]
        return MOCK_ALIGNMENT_RESPONSES.get(metric, {})
    elif "performance" in endpoint:
        metric = endpoint.split("/")[-1]
        return MOCK_PERFORMANCE_RESPONSES.get(metric, {})
    return {} 